from rsalor.utils.utils import is_convertable_to, memory_str, time_str, find_file
from rsalor.utils.CSV import CSV
from rsalor.utils.logger import Logger
from rsalor.utils.ali_to_fasta import ali_to_fasta



# Imports ----------------------------------------------------------------------
import os.path
from typing import List
import glob
import numpy as np
import ctypes


# Main -------------------------------------------------------------------------
def compute_weights(
        msa_path: str,
        msa_len: int,
        msa_depth: int,
        seqid: float=0.80,
        count_target_sequence: bool=True,
        num_threads: int=1,
        verboses: bool=False,
    ) -> List[float]:
    """Compute weights for all sequences of an MSA.
    Use C++ backend for time-performance. Implementation inspired from python package 'pycofitness'.
    
    Arguments:
        msa_path               (str):   path to msa '.fasta' file
        msa_len                (int):   length of the MSA (length of target sequence)
        msa_depth              (int):   depth of the MSA (number of sequences in the MSA)
        seqid                (float):   sequence identity threshold to consider two sequences as similar (default=0.80)
        count_target_sequence (bool):   count target sequence in weights computations
        num_threads            (int):   number of threads (CPUs) used by C++ backend (default=1)
        verboses              (bool):   set True to log steps of execution (default=False)
    
    Return:
        weights (List[float])
    """

    # Guardians
    assert msa_path.endswith(".fasta"), f"ERROR in compute_weights(): msa_path='{msa_path}' should end with '.fasta'."
    assert os.path.exists(msa_path), f"ERROR in compute_weights('{msa_path}'): msa_path='{msa_path}' files does not exist."
    assert 0.0 < seqid < 1.0, f"ERROR in compute_weights('{msa_path}'): seqid={seqid} (for clustering to compute weights) should be in [0, 1] excluded."
    assert num_threads > 0, f"ERROR in compute_weights('{msa_path}'): num_threads={num_threads} should be stricktly positive."

    # Find C++ computeWeightsBackend compiled executable file
    path_prefix = os.path.join(os.path.dirname(__file__), "lib_computeWeightsBackend*")
    backend_so_paths  = glob.glob(path_prefix)
    try:
        BACKEND_SO_PATH = backend_so_paths[0]
    except IndexError:
        error_log = "ERROR in compute_weights(): C++ computeWeightsBackend '.so' library path not found.\n"
        error_log += f"   * Unable to find C++ computeWeightsBackend '.so' library path in '{path_prefix}'\n"
        error_log += "   * Please install the pip package or compile the C++ code."
        raise ValueError(error_log)
    
    # Init C++ bridge
    computeWeightsBackend = ctypes.CDLL(BACKEND_SO_PATH)
    computeWeightsFunction = computeWeightsBackend.computeWeightsBackend
    computeWeightsFunction.argtypes = (
        ctypes.c_char_p,  # msa_path
        ctypes.c_uint,    # msa_len
        ctypes.c_uint,    # msa_depth
        ctypes.c_float,   # seqid
        ctypes.c_bool,    # count_target_sequence
        ctypes.c_uint,    # num_threads
        ctypes.c_bool     # verboses
    )
    computeWeightsFunction.restype = ctypes.POINTER(ctypes.c_float * msa_depth)
    freeWeights = computeWeightsBackend.freeWeights
    #freeWeights.argtypes # not need to define argtypes ???
    freeWeights.restype = None

    # Run backend
    weights_ptr = computeWeightsFunction(
        msa_path.encode('utf-8'),
        msa_len,
        msa_depth,
        seqid,
        count_target_sequence,
        num_threads,
        verboses,
    )

    # Convert to list
    weights = np.zeros((msa_depth), dtype=np.float32)
    for i, x in enumerate(weights_ptr.contents):
        weights[i]= x

    # Free memory
    weights_ptr_casted = ctypes.cast(weights_ptr, ctypes.POINTER(ctypes.c_void_p))
    freeWeights(weights_ptr_casted)

    # Return
    return weights


def write_weights(weights: List[float], weights_path: str) -> None:
    """Read weights list from a file."""

    # Guardians
    assert os.path.isdir(os.path.dirname(weights_path)), f"ERROR in write_weights(): directory of weights_path='{weights_path}' does not exist."
    assert len(weights) > 0, f"ERROR in write_weights(): weigths list can not be of length zero."

    # Write
    weights_str = "\n".join([str(w) for w in weights])
    with open(weights_path, "w") as fs:
        fs.write(weights_str)


def read_weights(weights_path: str) -> List[float]:
    """Write weights list to a file."""

    # Guardians (for input)
    assert os.path.isfile(weights_path), f"ERROR in read_weights(): weights_path='{weights_path}' file does not exist."

    # Read from file
    with open(weights_path, "r") as fs:
        lines = fs.readlines()

    # Parse
    weights: List[float] = []
    for i, line in enumerate(lines):
        if len(line) > 1 and line[0] != "#":
            try:
                weights.append(float(line))
            except:
                line = line.replace('\n', '')
                error_log = f"ERROR in read_weights(): failed to parse line {i+1} / {len(lines)} as a float."
                error_log += f" * weights_path='{weights_path}'"
                error_log += f" * line='{line}'"
                raise ValueError(error_log)
    if len(weights) == 0:
        raise ValueError(f"ERROR in read_weights(): no parsable weights line found in weights_path='{weights_path}'.")
    return weights



# Imports ----------------------------------------------------------------------
import os.path
from typing import Dict
import tempfile
import subprocess
from rsalor.utils import find_file
from rsalor.rsa.rsa_solver import RSASolver

# RSAMuSiC ---------------------------------------------------------------------
class RSAMuSiC(RSASolver):
    """
    RSAMuSiC(): Solver for RSA (Relative Solvent Accessibility) from a '.pdb' file using MuSiC software.
    
    usage:
        rsa_map = RSAMuSiC('./soft/MuSiC/music').run('./my_pdb.pdb')
    """

    # Constants ----------------------------------------------------------------

    # Saved condidate paths to simplify execution on different machines
    CANDIDATES_PATHS = [
        "music", "music_retro",                             # MuSiC executables
        "/home/Softs/MuSiC-4.1/music",                      # Nautilus and Santorin
    ]

    # Helper: hint to solve problem form the user: install software please
    HELPER_LOG = """-------------------------------------------------------
RSA Solver: MuSiC issue:
In order to solve Relative Solvent Accessiblity (RSA), RSALOR package uses MuSiC software.
MuSiC is our in house protein structure software (https://soft.dezyme.com/).
Please install MuSiC and specify the path to its executable or add it to system PATH.

Alternatively, if you do not have access to MuSiC, set rsa_solver='DSSP' and install DSSP (free for academic uses)
DSSP: https://swift.cmbi.umcn.nl/gv/dssp/

NOTE: you can still use the RSALOR package without MuSiC if you only want LOR values of the MSA without using RSA (just set pdb_path=None).
-------------------------------------------------------"""

    # Methods ------------------------------------------------------------------
    def __str__(self) -> str:
        return "RSASolver['MuSiC']"
    
    def execute_solver(self, pdb_path: str) -> Dict[str, float]:
        """Compute RSA by running MuSiC: 'music -cat'

        args:
        pdb_path (str):         path to PDB file   

        output:
        {resid: str => RSA: float}     (such as {'A13': 48.57, ...})
        """

        # Init MuSiC path (check MuSiC executable existance only if software is executed)
        self._init_music_path()

        # Using temporary directory
        with tempfile.TemporaryDirectory() as tmp_dir:

            # Init
            name = os.path.basename(pdb_path).removesuffix(".pdb")
            pdb_dir = os.path.abspath(os.path.dirname(pdb_path))
            path_in_path = os.path.join(tmp_dir, "path.in")
            cat_path = os.path.join(tmp_dir, f"{name}.cat")
            log_path = os.path.join(tmp_dir, f"log_{name}.txt")

            # Generate path.in file
            path_in = "\n".join([
                f"DATA    {os.path.join(os.path.dirname(self.music_path), 'MuSiC/Data/')}",
                f"PDB     {pdb_dir}/",
                f"OUTPUT  {tmp_dir}/",
                f"CAT     {tmp_dir}/\n"
            ])
            with open(path_in_path, "w") as fs:
                fs.write(path_in)

            # Adapt run to MuSiC 4.0 or 4.1 assuming version is specified in MuSiC folder name
            music_last_folder_name = os.path.basename(os.path.dirname(self.music_path))
            sidechain_parameter = "FULLATOM" if "4.1" in music_last_folder_name else ""

            # Run MuSiC cat
            music_cmd = f"{self.music_path} -cat {name} {sidechain_parameter} {name} -init {path_in_path} -log {name}"
            if self.verbose:
                print(f" * run MuSiC command: '{music_cmd}'")
            process = subprocess.run(music_cmd.split(), shell=False, capture_output=True, text=True)

            # Check execution errors
            if process.returncode != 0:
                self._log_music_execution_error(process, log_path, music_cmd)
                raise ValueError(f"ERROR in {self}.execute_solver(): execution of MuSiC failed.")
            if not os.path.isfile(cat_path):
                self._log_music_execution_error(process, log_path, music_cmd)
                raise ValueError(f"ERROR in {self}.execute_solver(): execution of MuSiC succeeded but no output '.cat' file is generated at '{cat_path}'.")
            
            # Log output
            #if self.verbose:
            #    print("\nMuSiC logs: ")
            #    music_lines = result.stdout.split("\n")
            #    music_lines = [line for line in music_lines if len(line) > 0]
            #    print("\n".join(music_lines) + "\n")

            # Parse MuSiC -cat output
            try:
                rsa_map = self._parse_cat(cat_path)
            except:
                raise ValueError(f"ERROR in {self}.execute_solver(): failed to parse RSA from generated file '{cat_path}'.")
            assert len(rsa_map) > 0, f"ERROR in {self}.execute_solver(): no valid RSA data found in file '{cat_path}'."

        # Return
        return rsa_map
    
    # Dependencies -------------------------------------------------------------
    def _init_music_path(self) -> str:
        """Find an existing executable file for MuSiC on the computer."""
        if self.executable_path is not None:
            music_path_list = [self.executable_path] + self.CANDIDATES_PATHS
        else:
            music_path_list = self.CANDIDATES_PATHS
        music_path = find_file(music_path_list, is_software=True, name="MuSiC", description=self.HELPER_LOG, verbose=self.verbose)
        self.music_path = music_path

    def _parse_cat(self, file_path: str) -> Dict[str, float]:
        """Parse music '.cat' file and return RSA mapping: {resid: str => RSA: float}."""

        # Guardians
        assert os.path.isfile(file_path), f"ERROR in {self}._parse_cat(): file_path='{file_path}' does not exists."
        assert file_path.endswith(".cat"), f"ERROR in {self}._parse_cat(): file_path='{file_path}' should end with '.cat'."

        # Read cat file
        with open(file_path, "r") as fs:

            # Skip lines before #RESIDUES section
            line = fs.readline()
            while line and not line.startswith("#RESIDUES"):
                if line.startswith("#RESIDUES"): break
                line = fs.readline()
            line = fs.readline()

            # Read #RESIDUES section
            rsa_map: Dict[str, float] = {}
            while line:
                
                # Break after #RESIDUES section ends
                if line.startswith("#"): break

                # Parse values from line
                resid = line[0:6].replace(" ", "")
                rsa = float(line[30:40])

                # Save values
                rsa_map[resid] = rsa
                line = fs.readline()

        # Sanity chech and return
        return rsa_map
    
    def _log_music_execution_error(self, music_run_process, log_path: str, music_cmd: str) -> None:
        print("\nERROR in MuSiC execution.")
        print(" * MuSiC command: ")
        print(f" $ {music_cmd}")
        print(" * Standard Output (stdout): ")
        print(music_run_process.stdout)
        print(" * Error Output (stderr): ")
        print(music_run_process.stderr)
        print(" * Log file content: ")
        if os.path.isfile(log_path):
            with open(log_path, "r") as fs:
                log_lines = "\n".join(list(fs.readlines()))
                print(log_lines)


"""
RSALOR
======

Combines structural data (Relative Solvent Accessibility, RSA) and
evolutionary data (Log Odds Ratio, LOR from MSA) to evaluate missense
mutations in proteins.

Example of usage in Python:

>>> from rsalor import MSA
>>> msa = MSA('./msa1.fasta', './pdb1.pdb', chain='A')
>>> scores = msa.get_scores()
>>> msa.save_scores('./msa1_rsalor-scores.csv')

Example of command line usage::

    $ rsalor ./msa1.fasta ./pdb1.pdb A -o ./msa1_rsalor-scores.csv
"""

from rsalor.msa import MSA




# Imports ----------------------------------------------------------------------
from typing import Union
from rsalor.sequence import AminoAcid

# Main -------------------------------------------------------------------------
class Residue:
    """Container class for a PDB residue.
    
    usage:
    res = Residue('A', '113', AminoAcid('K'))
    """

    # Constructor --------------------------------------------------------------
    def __init__(self, chain: str, position: str, amino_acid: AminoAcid, rsa: Union[None, float]=None):

        # Guardians
        assert len(chain) == 1 and chain != " ", f"ERROR in Residue(): invalid chain='{chain}'."
        if rsa is not None:
            assert rsa >= 0.0, f"ERROR in Residue(): rsa='{rsa}' should be positive."

        # Set properties
        self.chain: str = chain
        self.position: str = position
        self.amino_acid: AminoAcid = amino_acid
        self.rsa: Union[None, float] = rsa

    # Properties ---------------------------------------------------------------
    @property
    def resid(self) -> str:
        return self.chain + self.position

    def __str__(self) -> str:
        return f"Residue('{self.resid}', '{self.amino_acid.three}', RSA={self.rsa})"



# Imports ----------------------------------------------------------------------
import os.path
from typing import List, Tuple, Dict, Union
import Bio
from Bio.Align import PairwiseAligner
from rsalor.sequence import Sequence

# PairwiseAlignment ------------------------------------------------------------
class PairwiseAlignment:
    """Class to perform pairwise alignments based on sequence identity.
    The aim is to reconsile slightly different inputs of the same protein sequence but with possibly small incoherences like missing residues.
    Like the SEQRES and ATOM lines of a PDB.
    Or a sequence extracted from a PDB and a sequence from an MSA(for instance, MSA or PDB could cover a different range of the sequence).

    NOTE: Put sequence which is expected to contain the least gaps first
    
    usage:
    
    seq1 = FastaSequence("msa_seq", "HHALYDYEARTK") \n
    seq2 = FastaSequence("pdb_seq",   "ALYDYEART") \n
    align = PairwiseAlignment(seq1, seq2) \n
    align.show() \n
    seq1_to_seq2_id_mapping = align.get_mapping()
    """

    # Constants ----------------------------------------------------------------
    GAP_CHAR = "-"
    MATCH_CHAR = "|"
    MISMATCH_CHAR = "x"

    # Constructor --------------------------------------------------------------
    def __init__(
            self,
            sequence1: Sequence,
            sequence2: Sequence,
            match_score: float=1.0,
            mismatch_score: float=-3.0,
            open_gap_score: float=-2.5,
            extend_gap_score: float=-2.0,
            tail_gap_score: float=-2.0,
            query_insertion_multiplier: float=3.0,
        ):

        # Length Guardians
        if len(sequence1) == 0 or len(sequence2) == 0:
            print(f" * sequence1: {sequence1}")
            print(f" * sequence2: {sequence2}")
            raise ValueError("ERROR in PairwiseAlignment(): input target or query sequence can not be of length 0.")

        # Init base properties
        self.sequence1 = sequence1
        self.sequence2 = sequence2
        self.len1 = len(sequence1)
        self.len2 = len(sequence2)
        self.len_min = min(self.len1, self.len2)
        self.len_max = max(self.len1, self.len2)
        self.len_ratio = self.len_min / self.len_max

        # Init aligner
        self.aligner = PairwiseAligner()
        self.aligner.mode = 'global'
        self.aligner.match_score = match_score
        self.aligner.mismatch_score = mismatch_score
        # just because biopython decies to change its interface syntax each week ...
        if version_is_greater_than(Bio.__version__, "1.85"):
            self.aligner.open_internal_insertion_score = open_gap_score
            self.aligner.extend_internal_insertion_score = extend_gap_score
            self.aligner.open_right_insertion_score = tail_gap_score
            self.aligner.extend_right_insertion_score = tail_gap_score
            self.aligner.open_left_insertion_score = tail_gap_score
            self.aligner.extend_left_insertion_score = tail_gap_score
            self.aligner.open_internal_deletion_score = open_gap_score * query_insertion_multiplier
            self.aligner.extend_internal_deletion_score = extend_gap_score * query_insertion_multiplier
            self.aligner.open_left_deletion_score = tail_gap_score
            self.aligner.extend_left_deletion_score = tail_gap_score
            self.aligner.open_right_deletion_score = tail_gap_score
            self.aligner.extend_right_deletion_score = tail_gap_score
        else:
            self.aligner.target_internal_open_gap_score = open_gap_score
            self.aligner.target_internal_extend_gap_score = extend_gap_score
            self.aligner.target_right_open_gap_score = tail_gap_score
            self.aligner.target_right_extend_gap_score = tail_gap_score
            self.aligner.target_left_open_gap_score = tail_gap_score
            self.aligner.target_left_extend_gap_score = tail_gap_score
            self.aligner.query_internal_open_gap_score = open_gap_score * query_insertion_multiplier
            self.aligner.query_internal_extend_gap_score = extend_gap_score * query_insertion_multiplier
            self.aligner.query_left_open_gap_score = tail_gap_score
            self.aligner.query_left_extend_gap_score = tail_gap_score
            self.aligner.query_right_open_gap_score = tail_gap_score
            self.aligner.query_right_extend_gap_score = tail_gap_score

        # Align
        alignments = self.aligner.align(self.sequence1.sequence, self.sequence2.sequence)
        try: # For Biopython versions 1.80 and later
            self.align1: str = alignments[0][0]
            self.align2: str = alignments[0][1]
        except: # For legacy Biopython versions
            alignment_str_list = str(alignments[0]).split()
            self.align1: str = alignment_str_list[0]
            self.align2: str = alignment_str_list[2]
        self.score: float = alignments.score
        
        # Alignment properties
        self.match: int = 0
        self.gap: int = 0
        self.mismatch: int = 0
        comparator_list = []
        for aa1, aa2 in zip(self.align1, self.align2):
            if aa1 == self.GAP_CHAR or aa2 == self.GAP_CHAR:
                self.gap += 1
                comparator_list.append(self.GAP_CHAR)
            elif aa1 == aa2:
                self.match += 1
                comparator_list.append(self.MATCH_CHAR)
            else:
                self.mismatch += 1
                comparator_list.append(self.MISMATCH_CHAR)
        self.comparator = "".join(comparator_list)
        self.match_ratio: float = self.match / len(self)
        self.gap_ratio: float = self.gap / len(self)
        self.mismatch_ratio: float = self.mismatch / len(self)

        # Failed alignment error
        if self.match == 0:
            print("PairwiseAlignment(): failed to align sequences: zero matching positions is the alignemnt.")
            print(f" * sequence1: {sequence1}")
            print(f" * sequence2: {sequence2}")
            raise ValueError("ERROR in PairwiseAlignment(): alignment failed.")

        # Count gap types
        self.left_gap, self.right_gap = _count_tail_characters(self.comparator, self.GAP_CHAR)
        self.tail_gap: int = self.left_gap + self.right_gap
        self.internal_gap: int = self.gap - self.tail_gap

        # Count gap types by sequence
        self.gap1 = len(self) - self.len1
        self.gap2 = len(self) - self.len2
        self.left_gap1, self.right_gap1 = _count_tail_characters(self.align1, self.GAP_CHAR)
        self.tail_gap1: int = self.left_gap1 + self.right_gap1
        self.internal_gap1: int = self.gap1 - self.tail_gap1
        self.left_gap2, self.right_gap2 = _count_tail_characters(self.align2, self.GAP_CHAR)
        self.tail_gap2: int = self.left_gap2 + self.right_gap2
        self.internal_gap2: int = self.gap2 - self.tail_gap2

        # Some final measures
        self.sequence_identity: float = self.match / (self.match + self.mismatch) # excluding gapped positions
        self.coverage1: float = (self.match + self.mismatch) / self.len1
        self.coverage2: float = (self.match + self.mismatch) / self.len2
        self.coverage: float = (self.match + self.mismatch) / len(self)

    # Base Properties ----------------------------------------------------------
    def __len__(self) -> int:
        return len(self.align1)
    
    def __str__(self) -> str:
        return f"PairwiseAlignment('{self.sequence1.name}' vs. '{self.sequence2.name}', l={len(self)}, ({self.match} |, {self.gap} -, {self.mismatch} x))"
    
    def show(self, n_lines: int=120, only_critical_chunks: bool=False) -> "PairwiseAlignment":
        """Show the complete alignemnt."""
        assert n_lines > 0, f"ERROR in {self}.show(): n_lines={n_lines} should be > 0."
        print(self)
        l = len(self)
        i = 0
        while i < l:
            range_line = f"{i+1} - {min(i+n_lines, l)}"
            ali1_line = self.align1[i:i+n_lines]
            comp_line = self.comparator[i:i+n_lines]
            ali2_line = self.align2[i:i+n_lines]
            if only_critical_chunks:
                comp_line = comp_line.replace(self.MISMATCH_CHAR, f"\033[91m{self.MISMATCH_CHAR}\033[0m")
                ali2_line = ali2_line.replace(self.GAP_CHAR, f"\033[91m{self.GAP_CHAR}\033[0m")
            if not only_critical_chunks or self.MISMATCH_CHAR in comp_line or self.GAP_CHAR in ali2_line:
                print(range_line)
                print(ali1_line)
                print(comp_line)
                print(ali2_line)
            i += n_lines
        return self

    # Methods ------------------------------------------------------------------
    def write(self, save_path: str) -> "PairwiseAlignment":
        """Save alignment to a '.fasta' file."""
        save_path = os.path.abspath(save_path)
        assert save_path.endswith(".fasta"), f"ERROR in {self}.write(): save_path='{save_path}' sould be a '.fasta' file."
        assert os.path.isdir(os.path.dirname(save_path)), f"ERROR in {self}.write(): directory of save_path='{save_path}' does not exists."
        align_str = f">{self.sequence1.name}\n{self.align1}\n>{self.sequence2.name}\n{self.align2}\n"
        with open(save_path, "w") as fs:
            fs.write(align_str)
        return self

    def get_mapping(
            self,
            ids1: Union[None, List[Union[str, int]]]=None,
            ids2: Union[None, List[Union[str, int]]]=None,
            reversed: bool=False,
        ) -> Dict[Union[str, int], Union[str, int]]:
        """
        Return mapping of the aligment between mathing residues from seq1 to seq2.
            * By default the ids are just consecutive integers starting at 1.

        args:
            ids1: overwrite ids for seq1 (default is [1, 2, 3, ...])
            ids2: overwrite ids for seq2 (default is [1, 2, 3, ...])
            revered: if True, give mapping from seq2 to seq1
        """

        # Init ids
        if ids1 is None:
            ids1 = list(range(1, len(self.sequence1.sequence)+1))
        else:
            assert len(ids1) == len(self.sequence1), f"ERROR in {self}.get_mapping(): length of ids1={len(ids1)} does not match length of sequence1={len(self.sequence1)}."
        if ids2 is None:
            ids2 = list(range(1, len(self.sequence2.sequence)+1))
        else:
            assert len(ids2) == len(self.sequence2), f"ERROR in {self}.get_mapping(): length of ids2={len(ids2)} does not match length of sequence2={len(self.sequence2)}."

        # Manage reversed 
        align1, align2 = self.align1, self.align2
        if reversed:
            align1, align2 = align2, align1
            ids1, ids2 = ids2, ids1

        # Generate mapping
        mapping = {}
        i1, i2 = 0, 0
        for aa1, aa2 in zip(align1, align2):
            if aa1 != self.GAP_CHAR:
                i1 += 1
            if aa2 != self.GAP_CHAR:
                i2 += 1
            if aa1 != self.GAP_CHAR and aa2 != self.GAP_CHAR:
                mapping[ids1[i1-1]] = ids2[i2-1]
        return mapping
    
    @classmethod
    def get_gaps_ranges(cls, align: str, tail_gaps: bool=True) -> List[Tuple[int, int]]:
        """Return gaps ranges of alignment string."""
        
        # Detect gaps
        gaps_ranges = []
        is_previous_gap = False
        for i, aa in enumerate(align):
            is_current_gap = aa == cls.GAP_CHAR
            # Open gap range
            if not is_previous_gap and is_current_gap:
                current_gap_rang = [i]
            # Close gap range
            elif is_previous_gap and not is_current_gap:
                current_gap_rang.append(i)
                gaps_ranges.append(current_gap_rang)
            is_previous_gap = is_current_gap
        
        # Mange right tail gap
        if align[-1] == cls.GAP_CHAR:
            current_gap_rang.append(len(align))
            gaps_ranges.append(current_gap_rang)

        # Remove tail gaps if required
        if not tail_gaps:
            if align[0] == cls.GAP_CHAR:
                gaps_ranges = gaps_ranges[1:]
            if align[-1] == cls.GAP_CHAR:
                gaps_ranges = gaps_ranges[:-1]

        return gaps_ranges

# Dependencies -----------------------------------------------------------------
def _count_tail_characters(input_sequence: str, count_char: str) -> Tuple[int, int]:
    c1, c2 = 0, 0
    # Left tail
    for char in input_sequence:
        if char == count_char:
            c1 += 1
        else:
            break
    # Right tail
    for char in input_sequence[::-1]:
        if char == count_char:
           c2 += 1
        else:
            break
    return c1, c2

def version_is_greater_than(v1: str, v2: str, fallback_values: bool=False) -> bool:
    """Return if version v1 is greater than v2 or fallback_values if versions parsing has failed."""
    try:
        v1_list = [int(vi) for vi in v1.split(".")]
        v2_list = [int(vi) for vi in v2.split(".")]
        return v1_list > v2_list
    except:
        return fallback_values



from rsalor.sequence.amino_acid import AminoAcid
from rsalor.sequence.mutation import Mutation
from rsalor.sequence.sequence import Sequence
from rsalor.sequence.fasta_reader import FastaReader, FastaStream
from rsalor.sequence.pairwise_alignment import PairwiseAlignment




# Imports ----------------------------------------------------------------------
import os.path
from typing import List, Union
from rsalor.sequence import AminoAcid
from rsalor.sequence import Mutation

# Sequence ---------------------------------------------------------------------
class Sequence:
    """Container class for a single sequence (name, sequence and weight).
    
    usage:
    
    seq: Sequence = Sequence('seq1', 'MQIFVKTLTGKTI--T') \n
    seq_name: str = seq.name \n
    seq_str: str = seq.sequence \n
    seq.write('./fasta/seq1.fasta')
    """

    # Constants ----------------------------------------------------------------
    HEADER_START_CHAR = ">"
    GAP_CHAR = AminoAcid.GAP_ONE
    AMINO_ACIDS_IDENTITY_MAP = {aa.one: aa.one for aa in AminoAcid.get_all()} | {aa.one.lower(): aa.one.lower() for aa in AminoAcid.get_all()}

    # Constructor --------------------------------------------------------------
    def __init__(self, name: str, sequence: str, weight: float=1.0, to_upper: bool=True, convert_special_characters: bool=True):
        """Constructor for a (protein) Sequence object.
            name                         (str)         name of the sequence
            sequence                     (str)         amino acid sequence as a string
            weight                       (float=1.0)   weight of the sequence (in an MSA)
            to_upper                     (bool=True)   if True, convert all lower case amino acids to upper cases (such as in '.a2m' format)
            convert_special_characters   (bool=True)   if True, convert all non-standard characters (like '.' or '_') to a gap '-' (such as in '.a2m' format)
        """
        if name.startswith(self.HEADER_START_CHAR):
            name = name.removeprefix(self.HEADER_START_CHAR)
        if to_upper:
            sequence = sequence.upper()
        if convert_special_characters:
            gap = self.GAP_CHAR
            aa_map = self.AMINO_ACIDS_IDENTITY_MAP
            sequence = "".join([aa_map.get(aa, gap) for aa in sequence])
        self.name: str = name
        self.sequence: str = sequence
        self.weight: float = weight
    
    # Base properties ----------------------------------------------------------
    def __len__(self) -> int:
        return len(self.sequence)
    
    def __str__(self) -> str:
        MAX_PRINT_LEN = 15
        seq_str = self.sequence
        if len(seq_str) > MAX_PRINT_LEN:
            seq_str = f"{seq_str[0:MAX_PRINT_LEN]}..."
        name_str = self.name
        if len(name_str) > MAX_PRINT_LEN:
            name_str = f"{name_str[0:MAX_PRINT_LEN]}..."
        return f"Sequence('{name_str}', seq='{seq_str}', l={len(self)})"
    
    def __eq__(self, other: "Sequence") -> bool:
        return self.sequence == other.sequence
    
    def __neq__(self, other: "Sequence") -> bool:
        return self.sequence != other.sequence
    
    def __hash__(self) -> int:
        return hash(self.sequence)
    
    def __iter__(self):
        return iter(self.sequence)
    
    def __getitem__(self, id: int) -> str:
        return self.sequence[id]
    
    def __contains__(self, char: str) -> bool:
        return char in self.sequence
    
    # Base Methods -------------------------------------------------------------
    def n_gaps(self) -> int:
        """Return number of gaps in sequence."""
        return len([char for char in self.sequence if char == self.GAP_CHAR])
    
    def n_non_gaps(self) -> int:
        """Return number of non-gaps in sequence."""
        return len([char for char in self.sequence if char != self.GAP_CHAR])
    
    def gap_ratio(self) -> float:
        """Return gap ratio."""
        return self.n_gaps() / len(self)

    def contains_gaps(self) -> bool:
        """Return is sequence contains gaps."""
        for char in self.sequence:
            if char == self.GAP_CHAR:
                return True
        return False
    
    def is_all_amino_acids(self) -> bool:
        """Returns is sequence is composed of only standard amino acids."""
        for char in self.sequence:
            if not AminoAcid.one_exists(char):
                return False
        return True
    
    def to_fasta_string(self) -> str:
        """Return string of the sequence in FASTA format."""
        return f"{self.HEADER_START_CHAR}{self.name}\n{self.sequence}\n"
    
    def mutation_is_compatible(self, mutation: Union[str, Mutation]) -> bool:
        """Return if mutation is compatible with the sequence."""

        # Convert to Mutation type
        if isinstance(mutation, str):
            mutation = Mutation(mutation)

        # Verify if mutatoin position is in sequence
        if not (1 <= mutation.position <= len(self)):
            return False
        # Verify if wild-type amino acid corresponds to sequence
        if mutation.wt_aa.one != self.sequence[mutation.position-1]:
            return False
        return True

    # IO Methods ---------------------------------------------------------------
    def write(self, fasta_path: str) -> "Sequence":
        """Save sequence in a FASTA file."""

        # Guardians
        fasta_path = os.path.abspath(fasta_path)
        assert os.path.isdir(os.path.dirname(fasta_path)), f"ERROR in Sequence('{self.name}').write(): directory of '{fasta_path}' does not exists."
        assert fasta_path.endswith(".fasta"), f"ERROR in Sequence('{self.name}').write(): fasta_path='{fasta_path}' should end with '.fasta'."
        
        # Save FASTA and return self
        with open(fasta_path, "w") as fs:
            fs.write(self.to_fasta_string())
        return self

    # Mutate Methods -----------------------------------------------------------
    def trim(self, keep_positions: List[bool]) -> "Sequence":
        """Trim sequence (filter on positions) according to keep_positions (array of bool indicating which position to keep)."""

        # Guardians
        assert len(keep_positions) == len(self), f"ERROR in {self}.trim(): length of keep_positions ({len(keep_positions)}) does not match length of sequence ({len(self)})."

        # Trim and return self
        self.sequence = "".join([char for char, keep in zip(self.sequence, keep_positions) if keep])
        return self



# Imports ----------------------------------------------------------------------
import os.path
from typing import Union, List, Dict, Literal
from rsalor.sequence import AminoAcid
from rsalor.structure import Residue
from rsalor.sequence import Sequence
from rsalor.rsa import RSASolver, RSABiopython, RSADSSP, RSAMuSiC

# Execution --------------------------------------------------------------------
class Structure:
    """Structure object for parsing all Residues from ATOM lines and assign RSA (with biopython (Shrake & Rupley), DSSP or MuSiC).

    usage:
    structure = Structure('./my_pdb.pdb', 'A')
    """

    # Constants ----------------------------------------------------------------
    RSA_SOLVERS: Dict[str, RSASolver] = {
        "biopython": RSABiopython,
        "DSSP": RSADSSP,
        "MuSiC": RSAMuSiC,
    }

    # Constructor --------------------------------------------------------------
    def __init__(
            self,
            pdb_path: str,
            chain: str,
            rsa_solver: Literal["biopython", "DSSP", "MuSiC"]="biopython",
            rsa_solver_path: Union[None, str]=None,
            rsa_cache_path: Union[None, str]=None,
            verbose: bool=False,
        ):
        """Structure object for parsing all Residues from ATOM lines and assign RSA (with biopython, DSSP or MuSiC).

        arguments:
        pdb_path (str):                                   path to PDB file
        chain (str):                                      target chain in the PDB
        rsa_solver ('biopython'/'DSSP'/'MuSiC'):          solver to use to compute RSA
        rsa_solver_path (Union[None, str]=None):          path to solver executable
        rsa_cache_path (Union[None, str]=None):           path to write/read to/from RSA values
        verbose (bool=False):                             set True for logs
        """

        # Guardians
        assert os.path.isfile(pdb_path), f"ERROR in Structure(): pdb_path='{pdb_path}' file does not exist."
        assert pdb_path.endswith(".pdb"), f"ERROR in Structure(): pdb_path='{pdb_path}' should end with '.pdb'."
        assert len(chain) == 1 and chain != " ", f"ERROR in Structure(): chain='{chain}' should be a string of length 1 and not ' '."
        solver_list = list(self.RSA_SOLVERS.keys())
        assert rsa_solver in solver_list, f"ERROR in Structure(): rsa_solver='{rsa_solver}' should be in {solver_list}."

        # Init base properties
        self.pdb_path = pdb_path
        self.pdb_name = os.path.basename(self.pdb_path).removesuffix(".pdb")
        self.chain = chain
        self.name = f"{self.pdb_name}_{self.chain}"
        self.rsa_solver = rsa_solver
        self.rsa_solver_path = rsa_solver_path
        self.verbose = verbose

        # Parse structure
        self.residues: List[Residue] = []
        self.chain_residues: List[Residue] = []
        self.residues_map: Dict[str, Residue] = {}
        self._parse_structure()

        # Set sequence
        self.sequence = Sequence(f"{self.name} (PDB, ATOM-lines)", "".join(res.amino_acid.one for res in self.chain_residues))

        # Assign RSA
        solver: RSASolver = self.RSA_SOLVERS[rsa_solver]
        rsa_map = solver(self.rsa_solver_path, self.verbose).run(self.pdb_path, rsa_cache_path=rsa_cache_path)
        n_assigned_in_chain = 0
        for residue in self.residues:
            resid = residue.resid
            if resid in rsa_map:
                if residue.chain == self.chain:
                    n_assigned_in_chain += 1
                residue.rsa = rsa_map[resid]

        # Log
        if self.verbose:
            print(f" * {n_assigned_in_chain} / {len(self.chain_residues)} assigned RSA values for chain '{self.chain}'")


    # Base properties ----------------------------------------------------------
    def __str__(self) -> str:
        return f"Structure('{self.name}', l={len(self)})"

    def __len__(self) -> int:
        return len(self.residues)
    
    def __contains__(self, resid: str) -> bool:
        return resid in self.residues_map
    
    def __getitem__(self, id: int) -> dict:
        return self.residues[id]
    
    def __iter__(self):
        return iter(self.residues)
    
    # Deendencies --------------------------------------------------------------
    def _parse_structure(self) -> None:
        """Parse residues data from PDB file."""
        
        # Init
        model_counter = 0
        current_chain = None
        closed_chains = set()

        # Parse PDB residues
        with open(self.pdb_path, "r", encoding="ISO-8859-1") as fs:
            line = fs.readline()
            while line:
                prefix = line[0:6]
                
                # Atom line
                if prefix == "ATOM  " or prefix == "HETATM":
                    current_chain = line[21]
                    if current_chain in closed_chains: # discard ATOM line if chain is closed
                        line = fs.readline()
                        continue
                    position = line[22:27].replace(" ", "")
                    aa_three = line[17:20]
                    aa = AminoAcid.parse_three(aa_three)
                    if aa.is_unknown(): # discard non amino acid ATOM lines
                        line = fs.readline()
                        continue
                    resid = current_chain + position
                    if resid not in self.residues_map:
                        residue = Residue(current_chain, position, aa)
                        self.residues.append(residue)
                        self.residues_map[resid] = residue
                
                # Manage multiple models: consider only model 1
                elif prefix == "MODEL ":
                    model_counter += 1
                    if model_counter > 1:
                        #print(f"WARNING in {self}: PDB contains multiple models, but only model 1 will be considered.")
                        break

                # Manage closed chains: ATOMS that appears after the chain is closed are not part of the protein chain
                elif prefix == "TER   " or prefix == "TER\n":
                    if current_chain is not None:
                        closed_chains.add(current_chain)
                
                # Take next line
                line = fs.readline()
        
        # Set residues list of target chain
        self.chain_residues = [res for res in self.residues if res.chain == self.chain]

        # No target chain error
        if len(self.chain_residues) == 0:
            error_log = f"ERROR in {self}._parse_structure(): target chain '{self.chain}' not found in PDB file."
            error_log += f"\n * pdb_path: '{self.pdb_path}'"
            error_log += f"\n * num total residues: {len(self.residues)}"
            error_log += f"\n * existing chains: {list(set([res.chain for res in self.residues]))}"
            raise ValueError(error_log)



# Imports ----------------------------------------------------------------------
import os.path
import argparse
from rsalor import MSA


# CLI: dependencies ------------------------------------------------------------

# Custom Help Formatter (for 'rsalor -h' help logs menu)
class CustomHelpFormatter(argparse.HelpFormatter):

    # Increase max_help_position
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, max_help_position=40)

    # Only show metavar once
    def _format_action_invocation(self, action):
        if not action.option_strings:
            return super()._format_action_invocation(action)
        parts = action.option_strings[:]
        if action.nargs != 0:
            parts[-1] += f" {self._format_args(action, action.dest)}"
        return ", ".join(parts)


# CLI: main function -----------------------------------------------------------
def main():
    

    # Define Argument Parser ---------------------------------------------------

    # Init parser
    parser = argparse.ArgumentParser(
        description="Run RSALOR for all single-site missence mutations on a target protein sequence.",
        usage=f"rsalor <msa_path> <pdb_path> <chain> [options]\nhelp:  rsalor -h",
        formatter_class=CustomHelpFormatter,
    )

    parser.add_argument(
        "msa_path", type=str,
        help="path to MSA '.fasta' or '.a2m' file",
    )

    parser.add_argument(
        "pdb_path", type=str,
        help="path to PDB '.pdb' file",
    )

    parser.add_argument(
        "chain", type=str,
        help="chain in the PDB to consider",
    )

    parser.add_argument(
        "-o", "--output_path", type=str, default=None, metavar="<str>",
        help="path to output '.csv' file (default='./[msa_filename]_rsalor.csv')",
    )

    parser.add_argument(
        "-t", "--theta_regularization", type=float, default=0.01, metavar="<float>",
        help="regularization term at amino acid frequencies level (default=0.01)",
    )

    parser.add_argument(
        "-n", "--n_regularization", type=float, default=0.00, metavar="<float>",
        help="regularization term at amino acid counts level (default=0.00)",
    )

    parser.add_argument(
        "-i", "--ignore_target_sequence", dest="ignore_target_sequence", action="store_true",
        help="ignore target (first) sequence of the MSA in frequencies",
    )
    parser.set_defaults(ignore_target_sequence=False)

    parser.add_argument(
        "-r", "--keep_redundant_sequences", dest="keep_redundant_sequences", action="store_true",
        help="skip MSA pre-procesing step to remove redundent sequences",
    )
    parser.set_defaults(keep_redundant_sequences=False)

    parser.add_argument(
        "-w", "--seqid_weights", type=float, default=0.80, metavar="<float>",
        help="seqid threshold for weighting (default=0.80, set 0.0 to ignore)",
    )

    parser.add_argument(
        "-m", "--min_seqid", type=float, default=0.35, metavar="<float>",
        help="discard sequences which seqid with target sequence is below (default=0.35)",
    )

    parser.add_argument(
        "-N", "--num_threads", type=int, default=1, metavar="<int>",
        help="number of threads (CPUs) for weights step (default=1)",
    )

    parser.add_argument(
        "-S", "--silent", dest="silent", action="store_true",
        help="run in silent mode (default=False)",
    )
    parser.set_defaults(silent=False)

    parser.add_argument(
        "-W", "--disable_warnings", dest="disable_warnings", action="store_true",
        help="disable logging of warnings (default=False)",
    )
    parser.set_defaults(disable_warnings=False)

    parser.add_argument(
        "--sep", type=str, default=";", metavar="<str>",
        help="separator in the output '.csv' file",
    )

    parser.add_argument(
        "--weights_cache_path", type=str, default=None, metavar="<str>",
        help="set to read (if file exists) or write (if files does not exists) weights",
    )

    parser.add_argument(
        "--rsa_cache_path", type=str, default=None, metavar="<str>",
        help="set to read (if file exists) or write (if files does not exists) RSA values",
    )

    args = parser.parse_args()

    # Set default output_path
    if args.output_path is None:
        output_dir = "./"
        msa_name:str = os.path.basename(args.msa_path)
        for extention in MSA.ACCEPTED_EXTENTIONS:
            if msa_name.endswith(f".{extention}"):
                msa_name = msa_name.removesuffix(f".{extention}")
                break
        args.output_path = os.path.join(output_dir, f"{msa_name}_rsalor.csv")

    # Verify output_path
    else:
        if not args.output_path.endswith(".csv"):
            raise ValueError(f"ERROR in rsalor: output_path='{args.output_path}' should end with '.csv'.")
        
    # Set None values for 'degenerated' arguments
    if float(args.seqid_weights) == 0.0 or float(args.seqid_weights) == 1.0:
        args.seqid_weights = None
    if float(args.min_seqid) == 0.0:
        args.min_seqid = None


    # Execute RSALOR -----------------------------------------------------------

    # Run RSALOR
    msa = MSA(
        args.msa_path,
        args.pdb_path,
        args.chain,
        theta_regularization=args.theta_regularization,
        n_regularization=args.n_regularization,
        count_target_sequence=not args.ignore_target_sequence,
        remove_redundant_sequences=not args.keep_redundant_sequences,
        seqid_weights=args.seqid_weights,
        min_seqid=args.min_seqid,
        num_threads=args.num_threads,
        #rsa_solver=args.rsa_solver,
        #rsa_solver_path=args.rsa_solver_path,
        #trimmed_msa_path=args.trimmed_msa_path,
        #allow_msa_overwrite=args.allow_msa_overwrite,
        weights_cache_path=args.weights_cache_path,
        rsa_cache_path=args.rsa_cache_path,
        verbose=not args.silent,
        disable_warnings=args.disable_warnings,
        #name=args.name,
    )

    # Save scores
    msa.save_scores(
        args.output_path,
        round_digit=6,
        sep=args.sep,
    )




# Imports ----------------------------------------------------------------------
import os.path
import csv
from typing import Union, Tuple, List, Dict, Callable
import numpy as np


# Main -------------------------------------------------------------------------
class CSV:
    """
    Class to read/write a CSV file and manage it as a dataframe.
        * It never assumes a column or a cell type except when it is specified (all cells as :str by default).
        * Manages safety: impossible to have redundent values in header.
    """

    # Constants ----------------------------------------------------------------
    ALLOWED_EXTENTIONS = ["csv", "tsv"]

    # Constructor --------------------------------------------------------------
    def __init__(
            self,
            header: List[str]=[],
            sep: str=";",
            name: str="DataFrame",
            print_warnings: bool=True,
        ):

        # Base properties
        self.name = name
        self.print_warnings = print_warnings

        # Content
        self._header = Header(header, sep)
        self.entries = []

    # Basic properties ---------------------------------------------------------
    @property
    def sep(self) -> str:
        return self._header.sep

    def __len__(self) -> int:
        return len(self.entries)
    
    def __contains__(self, property_name: str) -> bool:
        return property_name in self._header
    
    def __getitem__(self, id: int) -> dict:
        return self.entries[id]
    
    def __iter__(self):
        return iter(self.entries)
    
    @property
    def n_rows(self) -> int:
        return len(self)
    
    @property
    def n_cols(self) -> int:
        return len(self._header)
    
    @property
    def shape(self) -> Tuple[int, int]:
        return (self.n_rows, self.n_cols)
    
    @property
    def df_size(self) -> int:
        return self.n_rows * self.n_cols
    
    def header(self) -> List[str]:
        return [p for p in self._header.properties]

    def __str__(self) -> str:
        return f"CSV('{self.name}', r={self.n_rows}, c={self.n_cols})"
    
    def warning(self, warning_str: str="") -> None:
        """Log a CSV Warning."""
        if self.print_warnings:
            print(f"WARNING in {self}{warning_str}")

    # Methods ------------------------------------------------------------------
    def set_sep(self, sep: str, safety_check: bool=True):
        """Set separator for the CSV (for .read and .write)"""

        if safety_check:
            
            # Computational time warning
            if self.df_size > 1000:
                self.warning(
                    f".set_sep('{sep}'): could be computationally expensive when CSV object already contains many entries. " + \
                    f"You can set 'safety_check' to False to skip coherence checks with separator."
                )

            # Guardians
            for entry in self.entries:
                for key, value in entry.items():
                    assert sep not in str(value), f"ERROR in {self}.set_sep('{sep}'): sep contained in entry's value ('{key}': '{value}')."
            
        # Set
        self._header.set_sep(sep)
        return self

    def add_entry(self, entry: dict):
        entry = {prop: entry[prop] for prop in self._header}
        self.entries.append(entry)
        return self

    def add_entries(self, entries: List[dict]):
        for entry in entries:
            self.add_entry(entry)
        return self

    def add_col(self, property: str, values: list, allow_replacement=False):
        if property in self._header:
            assert allow_replacement, f"ERROR in {self}.add_col(): property='{property}' already exists and allow_replacement is set to False."
        else:
            self._header.add(property)
        assert len(values) == len(self), f"ERROR in {self}.add_col(): values length ({len(values)}) != CSV length ({len(self)})."
        for entry, value in zip(self.entries, values):
            entry[property] = value
        return self
    
    def add_empty_col(self, property: str, missing_value: str="XXX", allow_replacement=False):
        values = [missing_value for _ in self.entries]
        self.add_col(property, values, allow_replacement=allow_replacement)
        return self
    
    def add_csv(self, other_csv):
        """Merge other_csv entries with current CSV (keeps header of current CSV)."""
        for property_name in self.header():
            assert property_name in other_csv._header, f"ERROR in {self}.add_csv(): property='{property_name}' does not exists in other_csv ({other_csv})."
        for entry in other_csv:
            self.add_entry(entry)
        return self

    def remove_col(self, property: str):
        self._header.remove(property)
        for entry in self.entries:
            del entry[property]
        return self

    def rename_col(self, property_old: str, property_new: str):
        self._header.rename(property_old, property_new)
        for entry in self.entries:
            entry[property_new] = entry[property_old]
            del entry[property_old]
        return self

    def order_header(self, header_order: List[str]):
        self._header.order(header_order)
        return self
    
    def filter(self, keep_entry_function: Callable, do_print: bool=False, filter_name: str=""):
        """Filter entries in the CSV with a filter_function."""
        l1 = len(self)
        self.entries = [entry for entry in self.entries if keep_entry_function(entry)]
        l2 = len(self)
        if do_print:
            print(f"{self}: Filter('{filter_name}'): {l1} -> {l2}")
        return self
    
    def set_col_type(self, property_name: str, dt: type, default_value=None):
        assert property_name in self._header, f"ERROR in {self}.set_col_type(): property_name='{property_name}' does not exists."
        for entry in self.entries:
            entry[property_name] = to_type(entry[property_name], dt, default_value=default_value)
    
    # Get Methods --------------------------------------------------------------
    def get_col(self, property: str, dt: Union[None, type]=None, default_value=None, as_numpy: bool=False):
        """Get Column of CSV as array."""
        assert property in self, f"ERROR in {self}.get_array('{property}'): property does not exists."
        col_list = [entry[property] for entry in self.entries]
        if dt is not None:
            col_list = [to_type(el, dt, default_value=default_value) for el in col_list]
        if as_numpy:
            col_list = np.array(col_list)
        return col_list
    
    def get_row(self, id: int, dt: Union[None, type]=None, default_value=None, as_numpy: bool=False):
        """Get Raw of CSV as array"""
        entry = self[id]
        row_list = [entry[p] for p in self._header]
        if dt is not None:
            row_list = [to_type(el, dt, default_value=default_value) for el in row_list]
        if as_numpy:
            row_list = np.array(row_list)
        return row_list
    
    def get_X(self, features: List[str]) -> np.ndarray:
        """Get features matrix X (numpy) from the CSV."""
        for feature in features:
            assert feature in self, f"ERROR in {self}.get_X(): feature='{feature}' does not exists."
        return np.array([
            [float(entry[feature]) for feature in features]
            for entry in self.entries
        ])
    
    def get_y(self, label: str) -> np.ndarray:
        """Get label array y (numpy) from the CSV."""
        assert label in self, f"ERROR in {self}.get_y(): label='{label}' does not exists."
        return np.array([float(entry[label]) for entry in self.entries])
    
    def get_Xy(self, features: List[str], label: str) -> Tuple[np.ndarray, np.ndarray]:
        """Get (features, label) tuple (X, y) (numpy) from the CSV."""
        return self.get_X(features), self.get_y(label)
    
    @staticmethod
    def hash_entry(entry: dict, hash_properties: List[str], sep: str="_") -> str:
        """Hash an entry (to :str) by values of its hash_properties."""
        return sep.join([entry[prop] for prop in hash_properties])
    
    @staticmethod
    def get_hash_entry(hash_properties: List[str], sep: str="_") -> Callable:
        """Generate a hash_entry function."""
        def hash_entry_function(entry: dict) -> str:
            return sep.join([entry[prop] for prop in hash_properties])
        return hash_entry_function

    def get_map(self, hash_properties: List[str], sep: str="_", map_function: Union[None, Callable]=None) -> Dict[str, Dict]:
        """
        Obtain a map {hash(entry) -> entry} from CSV (redundencies not allowed).
            * if map_function is set, values of the map are defined as map_function(entry)
        """
        for property in hash_properties:
            assert property in self, f"ERROR in {self}.to_map(): property='{property}' not in header."
        entries_map = {}
        for entry in self.entries:
            h = self.hash_entry(entry, hash_properties, sep=sep)
            assert h not in entries_map, f"ERROR in {self}.to_map({hash_properties}) redundency found for '{h}'."
            entries_map[h] = entry
        if map_function is not None:
            for h, entry in entries_map.items():
                entry[h] = map_function(entry)
        return entries_map

    def get_groups(self, hash_properties: List[str], sep: str="_", map_function: Union[None, Callable]=None) -> Dict[str, List[Dict]]:
        """
        Obtain a map for groups {hash(entry) -> [entries_list]} from CSV.
            * if map_function is set, values of the map are defined as [map_function(entry), ...]
        """
        for property in hash_properties:
            assert property in self, f"ERROR in {self}.to_map(): property='{property}' not in header."
        groups_map = {}
        for entry in self:
            h = self.hash_entry(entry, hash_properties, sep=sep)
            if h not in groups_map:
                groups_map[h] = []
            groups_map[h].append(entry)
        if map_function is not None:
            for h, group in groups_map.items():
                groups_map[h] = [map_function(e) for e in group]
        return groups_map
    
    def copy(self):
        """Copy CSV object."""
        new_csv = CSV()
        new_csv.name = self.name
        new_csv.print_warnings = self.print_warnings
        new_csv._header = self._header.copy()
        new_csv.entries = [
            {k: v for k, v in entry.items()}
            for entry in self.entries
        ]
        return new_csv
    
    def show(self, n_entries: int=5, min_colsize: int=3, max_colsize: int=20, max_linesize: int=200, round_digit: int=4, sep: str=" | ") -> None:
        """Show summary of CSV."""
        lines = [self._header.properties] + [self.get_row(id) for id in range(min(n_entries, len(self)))]
        col_sizes = [
            max([min_colsize, min([max([len(stringify_float(line[i], round_digit=round_digit)) for line in lines]), max_colsize])])
            for i in range(len(self._header))
        ]
        print(self)
        for line in lines:
            print_line(line, sizes=col_sizes, max_linesize=max_linesize, round_digit=round_digit, sep=sep)
        if len(self) > n_entries:
            print("  ...")
    
    # IO -----------------------------------------------------------------------
    def write(self, output_path: str):
        """Save to file."""

        # Guardians
        output_path = os.path.abspath(output_path)
        assert any([output_path.endswith(f".{extention}")] for extention in CSV.ALLOWED_EXTENTIONS), f"ERROR in {self}.write('{output_path}'): extention sould be among {CSV.ALLOWED_EXTENTIONS})."
        assert os.path.isdir(os.path.dirname(output_path)), f"ERROR in {self}.write('{output_path}'): destination folder does not exists."
        if output_path.endswith("tsv"):
            assert self.sep == "\t",  f"ERROR in {self}.write('{output_path}'): if extention is '.tsv', separator should be '\\t' however sep='{self.sep}'."
        
        # Stringify
        str_header = self.sep.join(self._header.properties)
        str_entries_list = [
            self.sep.join(str(entry[prop]) for prop in self._header.properties)
            for entry in self.entries
        ]
        str_lines = [str_header] + str_entries_list

        # Write
        with open(output_path, "w") as fs:
            fs.write("\n".join(str_lines))
        return self
    
    def read(self, input_path: str, col_types: Dict[str, type]={}, col_default: dict={}):
        """Read from file."""

        # Guardians
        assert any([input_path.endswith(f".{extention}")] for extention in CSV.ALLOWED_EXTENTIONS), f"ERROR in {self}.read('{input_path}'): extention sould be among {CSV.ALLOWED_EXTENTIONS})."
        assert os.path.isfile(input_path), f"ERROR in {self}.read('{input_path}'): input_path file does not exists."

        # Set name
        file_name = os.path.basename(input_path)
        name = ".".join(file_name.split(".")[:-1])
        self.name = name

        # Parse csv from file
        with open(input_path, newline='') as csvfile:
            csv_lines = list(csv.reader(csvfile, delimiter=self.sep))

        # Set CSV header
        header = csv_lines[0]
        if len(header) <= 1:
            self.warning(f".read('{input_path}'): header contains {len(header)} values. Maybe sep='{self.sep}' parameter in incorrect.")
        self._header = Header(header, self.sep)

        # Set CSV entries
        self.entries = []
        for i, line in enumerate(csv_lines[1:]):
            assert len(line) == len(header), f"ERROR in {self}.read('{input_path}'): number of elements ({len(line)}) in entry ({i+1}/{len(csv_lines)-1}) does not match the header ({len(header)})."
            self.entries.append({prop: value for prop, value in zip(header, line)})

        # Set column types if required
        for col_name, dt in col_types.items():
            col_default_value = col_default.get(col_name, None)
            self.set_col_type(col_name, dt, default_value=col_default_value)

        return self

# Dependencies -----------------------------------------------------------------

class Header:
    """
    Container for the Header of a CSV object.
        -> ordered list with no repetitions allowed and a separator of length = 1.
    """

    # Constructor --------------------------------------------------------------
    def __init__(self, properties: List[str], sep: str):

        # Init
        self.sep = ""
        self.properties = []
        self.properties_set = set()

        # Set header values
        self.set_sep(sep)
        for property in properties:
            self.add(property)

    # Basic properties ---------------------------------------------------------
    def __getitem__(self, id: int) -> str:
        return self.properties[id]

    def __iter__(self):
        return iter(self.properties)
    
    def __contains__(self, property_name: str) -> bool:
        return property_name in self.properties_set

    def __len__(self) -> int:
        return len(self.properties)
    
    def __str__(self) -> str:
        return f"CSV.Header(l={len(self)})"

    def show(self):
        MAX_CHAR = 80
        properties_str = f"'{self.properties[0]}'"
        for property in self.properties[1:]:
            if len(properties_str) + len(property) > MAX_CHAR:
                properties_str += ", ..."
                break
            properties_str += f", '{property}'"
        print(f"CSV.Header([{properties_str}], len={len(self)}, sep='{self.sep}')")
        return self
    
    def idof(self, property_name: str) -> int:
        assert property_name in self, f"ERROR in {self}.idof(): property_name='{property_name}' not in header."
        for i, current_property_name in enumerate(self):
            if property_name == current_property_name:
                return i
    
    # Methods ------------------------------------------------------------------
    def set_sep(self, sep: str):
        assert len(sep) == 1, f"ERROR in {self}.set_sep(): sep='{sep}' should be of length 1."
        for property in self:
            assert sep not in property, f"ERROR in {self}.set_sep(): sep='{sep}' is contained in property '{property}'."        
        self.sep = sep
        return self

    def add(self, property_name: str):
        assert self.sep not in property_name, f"ERROR in {self}.add('{property_name}'): property contains sep='{self.sep}'."
        assert property_name not in self, f"ERROR in {self}.add('{property_name}'): property already exists."
        self.properties.append(property_name)
        self.properties_set.add(property_name)
        return self

    def remove(self, property_name: str):
        assert property_name in self, f"ERROR in {self}.remove('{property_name}'): property does not exists."
        self.properties.remove(property_name)
        self.properties_set.remove(property_name)
        return self

    def rename(self, property_old: str, property_new: str):
        assert property_old != property_new, f"ERROR in {self}.rename(): old property and new property have the same value '{property_old}'."
        assert property_old in self, f"ERROR in {self}.rename(): old property '{property_old}' is not in header."
        assert property_new not in self, f"ERROR in {self}.rename(): new property '{property_new}' already in header."
        assert self.sep not in property_new, f"ERROR in {self}.rename(): new property '{property_new}' contains sep='{self.sep}'."
        id = self.idof(property_old)
        self.properties[id] = property_new
        self.properties_set.add(property_new)
        self.properties_set.remove(property_old)
        return self

    def order(self, header_order: List[str]):
        for property in header_order:
            assert property in self, f"ERROR in {self}.order(): property '{property}' not in header."
        ordered_properties_set = set(header_order)
        unordered_properties = [property for property in self if property not in ordered_properties_set]
        self.properties = header_order + unordered_properties
        return self
    
    def copy(self):
        return Header([p for p in self], self.sep)
    
# Dependency: Utils Funcions ---------------------------------------------------
def to_type(input, dt:type, default_value=None):
    """Convert input to type dt. If default_value is set, returns default_value when convertion fails."""
    try:
        return dt(input)
    except:
        if default_value is None:
            raise ValueError(f"ERROR in CSV().to_type(): input='{input}' not convertable to {dt}. Please correct input or set a default_value.")
        else:
            return default_value

def print_line(
        line_list,
        sep: str=" | ", dots_str: str="...",
        size: int=20, sizes: Union[None, List[int]]=None, max_linesize: int=200,
        round_digit: int=4,
    ) -> None:
    """Print a line from a table (dataframe) in a standardized way."""
    if sizes is None: sizes = [size for _ in line_list]
    line_str = ""
    unprinted_cols = False
    for element, size in zip(line_list, sizes):
        line_new_col = sep + format_string(element, size, round_digit=round_digit)
        if len(line_str) + len(line_new_col) > max_linesize - (len(sep) + len(dots_str)):
            unprinted_cols = True
            break
        line_str += line_new_col
    if unprinted_cols:
        line_str += sep + "..."
    line_str += sep
    print(line_str[1:-1])

def format_string(input, size: int=20, filler: int=" ", dots_str: str="...", round_digit: int=4) -> str:
    """Format a string to standardized form (length, ...)"""
    input_str = stringify_float(input, round_digit=round_digit)
    if len(input_str) > size:
        return input_str[:size-len(dots_str)] + dots_str
    else:
        return input_str + filler*(size - len(input_str))
    
def stringify_float(input, round_digit: int=4) -> str:
    if isinstance(input, float):
        str_float = str(round(input, round_digit))
        n_digits = len(str_float.split(".")[-1])
        str_float  = str_float + ("0"*(round_digit-n_digits))
        if str_float[0] != "-":
            str_float = " " + str_float
        return str_float
    else:
        return str(input)




# Imports ----------------------------------------------------------------------
from typing import List, Dict

# AminoAcid --------------------------------------------------------------------
class AminoAcid:
    """Container class for the 20 standard proteogenic amino acids.
        * Manages mapping between amino acids: id, one-letter-code and three-letter-code
        * Manages non-standard amino acids three-letters-codes and their corresponding standard AA

    usage:
    
    alanine_id: int = AminoAcid.ONE_2_ID['A'] \n
    one_letter_code_3: str = AminoAcid.ID_2_ONE[3] \n
    ala: AminoAcid= AminoAcid('A') \n
    non_standard_metionine: AminoAcid = AminoAcid.parse_three('MSE')
    """

    # Static Properties --------------------------------------------------------

    # Standard Amino Acids metadata
    _AA_LIST = [
        ( 0, "A", "ALA", "alanine"),
        ( 1, "C", "CYS", "cysteine"),
        ( 2, "D", "ASP", "aspartate"),
        ( 3, "E", "GLU", "glutamate"),
        ( 4, "F", "PHE", "phenylalanine"),
        ( 5, "G", "GLY", "glycine"),
        ( 6, "H", "HIS", "histidine"),
        ( 7, "I", "ILE", "isoleucine"),
        ( 8, "K", "LYS", "lysine"),
        ( 9, "L", "LEU", "leucine"),
        (10, "M", "MET", "methionine"),
        (11, "N", "ASN", "asparagine"),
        (12, "P", "PRO", "proline"),
        (13, "Q", "GLN", "glutamine"),
        (14, "R", "ARG", "arginine"),
        (15, "S", "SER", "serine"),
        (16, "T", "THR", "thréonine"),
        (17, "V", "VAL", "valine"),
        (18, "W", "TRP", "tryptophane"),
        (19, "Y", "TYR", "tyrosine"),
    ]

    # Gap and Unknown Amino Acid properties
    GAP_ID = 20
    GAP_ONE = "-"
    GAP_THREE = "GAP"
    GAP_NAME = "gap"
    UNK_ID = -1
    UNK_ONE = "X"
    UNK_THREE = "XXX"
    UNK_NAME = "unknown"

    # Non-Standard Amino Acids three-letter-codes mapped to closest Standard Amino Acids
    _NON_STANDARD_AAS = {
        # WARNING: We do not represent here ambiguous mappings like:   "GLX" => "GLN" or "GLU"
        "4HT": "TRP", "CLG": "LYS", "HSE": "SER", "BIF": "PHE", "B3D": "ASP", "BB8": "PHE", "3MY": "TYR", "SNK": "HIS",
        "3CF": "PHE", "A5N": "ASN", "LED": "LEU", "TOX": "TRP", "CR5": "GLY", "ILM": "ILE", "0A9": "PHE", "DAS": "ASP",
        "NYS": "CYS", "73P": "LYS", "MSO": "MET", "IYR": "TYR", "PR9": "PRO", "R4K": "TRP", "L5P": "LYS", "31Q": "CYS",
        "OCY": "CYS", "BH2": "ASP", "XSN": "ASN", "SXE": "SER", "GMA": "GLU", "SEP": "SER", "CYD": "CYS", "YPZ": "TYR",
        "GPL": "LYS", "RVX": "SER", "YCM": "CYS", "SEL": "SER", "DNE": "LEU", "LEN": "LEU", "4FB": "PRO", "4OU": "PHE",
        "LGY": "LYS", "TTQ": "TRP", "DBB": "THR", "LBZ": "LYS", "QX7": "ALA", "H14": "PHE", "CIR": "ARG", "73O": "TYR",
        "EI4": "ARG", "LVN": "VAL", "SRZ": "SER", "55I": "PHE", "UF0": "SER", "YHA": "LYS", "QM8": "ALA", "TQQ": "TRP",
        "QIL": "ILE", "Q75": "MET", "11Q": "PRO", "A8E": "VAL", "DHV": "VAL", "3BY": "PRO", "2ZC": "SER", "T9E": "THR",
        "CSZ": "CYS", "5CS": "CYS", "KPI": "LYS", "0AH": "SER", "HSK": "HIS", "TH6": "THR", "ARO": "ARG", "E9V": "HIS",
        "UXQ": "PHE", "MHL": "LEU", "CAS": "CYS", "8RE": "LYS", "LLP": "LYS", "PTH": "TYR", "ORQ": "ARG", "73N": "ARG",
        "BTK": "LYS", "HVA": "VAL", "LMQ": "GLN", "FME": "MET", "XX1": "LYS", "I7F": "SER", "4N9": "PRO", "TYJ": "TYR",
        "BOR": "ARG", "HL2": "LEU", "73C": "SER", "0CS": "ALA", "AGM": "ARG", "CYW": "CYS", "ASL": "ASP", "I3D": "TRP",
        "NPH": "CYS", "JKH": "PRO", "QMB": "ALA", "XCN": "CYS", "PHI": "PHE", "NAL": "ALA", "LYZ": "LYS", "6M6": "CYS",
        "VAD": "VAL", "EXL": "TRP", "WFP": "PHE", "823": "ASN", "CLH": "LYS", "C6C": "CYS", "DCY": "CYS", "DPP": "ALA",
        "KHB": "LYS", "DNW": "ALA", "BUC": "CYS", "CSU": "CYS", "H5M": "PRO", "RXL": "VAL", "FOE": "CYS", "GHP": "GLY",
        "2KP": "LYS", "OMX": "TYR", "ZCL": "PHE", "MGG": "ARG", "DLS": "LYS", "30V": "CYS", "02K": "ALA", "DA2": "ARG",
        "TYY": "TYR", "HRG": "ARG", "PHL": "PHE", "PRJ": "PRO", "M2L": "LYS", "SUN": "SER", "TSY": "CYS", "PF5": "PHE",
        "4CF": "PHE", "1OP": "TYR", "CSB": "CYS", "POM": "PRO", "ELY": "LYS", "TRQ": "TRP", "BP5": "ALA", "5VV": "ASN",
        "6DN": "LYS", "MIS": "SER", "MLZ": "LYS", "EME": "GLU", "4J5": "ARG", "MPQ": "GLY", "LLO": "LYS", "FQA": "LYS",
        "PR7": "PRO", "NLW": "LEU", "OMY": "TYR", "5CT": "LYS", "PRK": "LYS", "DPQ": "TYR", "N0A": "PHE", "3QN": "LYS",
        "K5H": "CYS", "HNC": "CYS", "TYO": "TYR", "Q3P": "LYS", "BWV": "ARG", "4L0": "PRO", "ZAL": "ALA", "IAM": "ALA",
        "AGQ": "TYR", "07O": "CYS", "PCA": "GLN", "2MR": "ARG", "TRN": "TRP", "4AR": "ARG", "HLY": "LYS", "DHI": "HIS",
        "J2F": "TYR", "C3Y": "CYS", "GL3": "GLY", "BTR": "TRP", "OYL": "HIS", "IGL": "GLY", "2GX": "PHE", "8LJ": "PRO",
        "AYA": "ALA", "XYC": "ALA", "CY1": "CYS", "CGU": "GLU", "PM3": "PHE", "03Y": "CYS", "CE7": "ASN", "HSL": "SER",
        "BXT": "SER", "MHU": "PHE", "HOX": "PHE", "5GM": "ILE", "DVA": "VAL", "CYR": "CYS", "YOF": "TYR", "DDZ": "ALA",
        "4PQ": "TRP", "ECC": "GLN", "GHG": "GLN", "IPG": "GLY", "PPN": "PHE", "L3O": "LEU", "AEA": "CYS", "7N8": "PHE",
        "AHO": "ALA", "TBG": "VAL", "BFD": "ASP", "HPE": "PHE", "5MW": "LYS", "U2X": "TYR", "N10": "SER", "TGH": "TRP",
        "51T": "TYR", "DDE": "HIS", "DBZ": "ALA", "FF9": "LYS", "HTN": "ASN", "NVA": "VAL", "HS9": "HIS", "ACB": "ASP",
        "9KP": "LYS", "FTR": "TRP", "ALS": "ALA", "DYJ": "PRO", "RPI": "ARG", "FTY": "TYR", "TQZ": "CYS", "FVA": "VAL",
        "CS4": "CYS", "QVA": "CYS", "XPR": "PRO", "0QL": "CYS", "TCQ": "TYR", "OXX": "ASP", "ZZJ": "ALA", "LDH": "LYS",
        "3CT": "TYR", "H7V": "ALA", "4N7": "PRO", "PYA": "ALA", "WVL": "VAL", "DMK": "ASP", "EFC": "CYS", "0BN": "PHE",
        "MHO": "MET", "ECX": "CYS", "ESB": "TYR", "KGC": "LYS", "3WX": "PRO", "MBQ": "TYR", "ILX": "ILE", "DSG": "ASN",
        "P2Q": "TYR", "LSO": "LYS", "6CW": "TRP", "SDP": "SER", "MP8": "PRO", "HTR": "TRP", "B3S": "SER", "TYB": "TYR",
        "PAQ": "TYR", "HS8": "HIS", "RX9": "ILE", "DHA": "SER", "CHP": "GLY", "MMO": "ARG", "FCL": "PHE", "05O": "TYR",
        "ICY": "CYS", "DIV": "VAL", "N65": "LYS", "Q78": "PHE", "KCR": "LYS", "TY8": "TYR", "GVL": "SER", "MLL": "LEU",
        "DNP": "ALA", "5XU": "ALA", "O7D": "TRP", "NFA": "PHE", "DBY": "TYR", "QCS": "CYS", "ZYK": "PRO", "IIL": "ILE",
        "ABA": "ALA", "4AW": "TRP", "BSE": "SER", "LLY": "LYS", "4D4": "ARG", "MNL": "LEU", "FGL": "GLY", "SET": "SER",
        "MYN": "ARG", "C4R": "CYS", "CZZ": "CYS", "CZS": "ALA", "Y1V": "LEU", "CWR": "SER", "NBQ": "TYR", "KYQ": "LYS",
        "2TY": "TYR", "1PA": "PHE", "6V1": "CYS", "FGP": "SER", "BB9": "CYS", "AGT": "CYS", "CYG": "CYS", "VI3": "CYS",
        "PH6": "PRO", "NZH": "HIS", "DAB": "ALA", "B2A": "ALA", "6WK": "CYS", "PR4": "PRO", "7O5": "ALA", "OHS": "ASP",
        "3YM": "TYR", "Z3E": "THR", "NC1": "SER", "CAF": "CYS", "BPE": "CYS", "BB7": "CYS", "RE0": "TRP", "TSQ": "PHE",
        "4CY": "MET", "G5G": "LEU", "TDD": "LEU", "KCX": "LYS", "0AR": "ARG", "HSV": "HIS", "2ML": "LEU", "4PH": "PHE",
        "V44": "CYS", "IAS": "ASP", "FH7": "LYS", "PTM": "TYR", "SAR": "GLY", "SVX": "SER", "MEN": "ASN", "CS1": "CYS",
        "HOO": "HIS", "NYB": "CYS", "HMR": "ARG", "05N": "PRO", "V61": "PHE", "41H": "PHE", "BMT": "THR", "4HL": "TYR",
        "I2M": "ILE", "4N8": "PRO", "2RX": "SER", "CS3": "CYS", "MEA": "PHE", "B2F": "PHE", "CYF": "CYS", "GNC": "GLN",
        "4HJ": "SER", "CSJ": "CYS", "2SO": "HIS", "Q2E": "TRP", "CXM": "MET", "4WQ": "ALA", "5OW": "LYS", "TRX": "TRP",
        "B3Y": "TYR", "DAH": "PHE", "5PG": "GLY", "ESC": "MET", "DTY": "TYR", "CGA": "GLU", "TFW": "TRP", "SMF": "PHE",
        "S1H": "SER", "SAC": "SER", "QCI": "GLN", "CMT": "CYS", "TY2": "TYR", "0A8": "CYS", "OMH": "SER", "QPA": "CYS",
        "MK8": "LEU", "DLE": "LEU", "T0I": "TYR", "ALT": "ALA", "3X9": "CYS", "5CW": "TRP", "9E7": "LYS", "MGN": "GLN",
        "PBF": "PHE", "AEI": "THR", "TYI": "TYR", "SNN": "ASN", "74P": "LYS", "OHI": "HIS", "KST": "LYS", "SBL": "SER",
        "JJJ": "CYS", "JJL": "CYS", "2RA": "ALA", "DIL": "ILE", "02Y": "ALA", "CYJ": "LYS", "2HF": "HIS", "FC0": "PHE",
        "NLN": "LEU", "XW1": "ALA", "QMM": "GLN", "TOQ": "TRP", "WPA": "PHE", "TIH": "ALA", "NLB": "LEU", "BG1": "SER",
        "PTR": "TYR", "0WZ": "TYR", "ZYJ": "PRO", "SNC": "CYS", "BBC": "CYS", "B3E": "GLU", "4GJ": "CYS", "MSA": "GLY",
        "TPO": "THR", "HIQ": "HIS", "PHA": "PHE", "THC": "THR", "JJK": "CYS", "API": "LYS", "TY5": "TYR", "LPD": "PRO",
        "MND": "ASN", "PRV": "GLY", "M3L": "LYS", "HR7": "ARG", "86N": "GLU", "DSN": "SER", "5R5": "SER", "IC0": "GLY",
        "ARM": "ARG", "4AK": "LYS", "HT7": "TRP", "E9M": "TRP", "4DP": "TRP", "IML": "ILE", "BCS": "CYS", "7OZ": "ALA",
        "2MT": "PRO", "GLZ": "GLY", "0E5": "THR", "U3X": "PHE", "HYP": "PRO", "M0H": "CYS", "7XC": "PHE", "AZK": "LYS",
        "AHB": "ASN", "NCB": "ALA", "ASA": "ASP", "TPL": "TRP", "0TD": "ASP", "HTI": "CYS", "LRK": "LYS", "ME0": "MET",
        "143": "CYS", "FY2": "TYR", "1TY": "TYR", "QPH": "PHE", "F2F": "PHE", "3PX": "PRO", "PLJ": "PRO", "N9P": "ALA",
        "3ZH": "HIS", "C5C": "CYS", "PFF": "PHE", "NEP": "HIS", "CSA": "CYS", "4J4": "CYS", "O7G": "VAL", "TTS": "TYR",
        "KFP": "LYS", "FZN": "LYS", "TYN": "TYR", "AA4": "ALA", "LYX": "LYS", "HP9": "PHE", "TH5": "THR", "D2T": "ASP",
        "MED": "MET", "TRW": "TRP", "HLU": "LEU", "CSO": "CYS", "23F": "PHE", "PG9": "GLY", "EJA": "CYS", "RE3": "TRP",
        "66D": "ILE", "4OG": "TRP", "MSE": "MET", "MDF": "TYR", "DBU": "THR", "SEN": "SER", "Y57": "LYS", "XA6": "PHE",
        "M2S": "MET", "FLT": "TYR", "GME": "GLU", "LE1": "VAL", "FY3": "TYR", "OZW": "PHE", "FP9": "PRO", "FHL": "LYS",
        "MLE": "LEU", "DAR": "ARG", "BHD": "ASP", "LA2": "LYS", "SLZ": "LYS", "CSX": "CYS", "OCS": "CYS", "DMH": "ASN",
        "2CO": "CYS", "NLE": "LEU", "LME": "GLU", "HIC": "HIS", "ZBZ": "CYS", "MYK": "LYS", "2JG": "SER", "ORN": "ALA",
        "YTF": "GLN", "1AC": "ALA", "OLD": "HIS", "B2I": "ILE", "HZP": "PRO", "4AF": "PHE", "OMT": "MET", "CSP": "CYS",
        "APK": "LYS", "DPR": "PRO", "CY0": "CYS", "5T3": "LYS", "CY3": "CYS", "3GL": "GLU", "4II": "PHE", "0AK": "ASP",
        "ALC": "ALA", "LP6": "LYS", "HIP": "HIS", "60F": "CYS", "CML": "CYS", "CYQ": "CYS", "NA8": "ALA", "MH6": "SER",
        "GFT": "SER", "WLU": "LEU", "AZH": "ALA", "KBE": "LYS", "LCK": "LYS", "LAY": "LEU", "0LF": "PRO", "KKD": "ASP",
        "K7K": "SER", "CSR": "CYS", "B3K": "LYS", "OSE": "SER", "F2Y": "TYR", "NMM": "ARG", "P1L": "CYS", "PRS": "PRO",
        "OBS": "LYS", "ZDJ": "TYR", "BYR": "TYR", "HY3": "PRO", "ASB": "ASP", "NLY": "GLY", "0A1": "TYR", "DPL": "PRO",
        "SCS": "CYS", "I4G": "GLY", "6CV": "ALA", "HIA": "HIS", "LYN": "LYS", "54C": "TRP", "FGA": "GLU", "B27": "THR",
        "TYE": "TYR", "DTH": "THR", "PSH": "HIS", "EXA": "LYS", "BLE": "LEU", "P9S": "CYS", "23P": "ALA", "1TQ": "TRP",
        "RVJ": "ALA", "ALO": "THR", "FL6": "ASP", "4LZ": "TYR", "TMD": "THR", "FHO": "LYS", "0FL": "ALA", "AN6": "LEU",
        "4OV": "SER", "432": "SER", "SCH": "CYS", "DGL": "GLU", "2TL": "THR", "TPQ": "TYR", "3AH": "HIS", "CSD": "CYS",
        "PR3": "CYS", "IZO": "MET", "DV9": "GLU", "41Q": "ASN", "DI7": "TYR", "34E": "VAL", "MHS": "HIS", "GGL": "GLU",
        "ALY": "LYS", "O6H": "TRP", "8JB": "CYS", "SVV": "SER", "KOR": "MET", "PYX": "CYS", "6CL": "LYS", "WRP": "TRP",
        "SCY": "CYS", "G1X": "TYR", "2KK": "LYS", "TYQ": "TYR", "MIR": "SER", "ALN": "ALA", "CMH": "CYS", "KPY": "LYS",
        "SVZ": "SER", "NMC": "GLY", "RGL": "ARG", "SME": "MET", "DAL": "ALA", "DTR": "TRP", "PEC": "CYS", "SGB": "SER",
        "NLO": "LEU", "AHP": "ALA", "SLL": "LYS", "TRF": "TRP", "CME": "CYS", "SEE": "SER", "MME": "MET", "DYA": "ASP",
        "33X": "ALA", "LYF": "LYS", "CZ2": "CYS", "TRO": "TRP", "DPN": "PHE", "IB9": "TYR", "POK": "ARG", "LET": "LYS",
        "CCS": "CYS", "DGN": "GLN", "NIY": "TYR", "E9C": "TYR", "SEB": "SER", "AIB": "ALA", "OAS": "SER", "V7T": "LYS",
        "K5L": "SER", "TYS": "TYR", "FIO": "ARG", "B2V": "VAL", "GLJ": "GLU", "JLP": "LYS", "MVA": "VAL", "0Y8": "PRO",
        "OTH": "THR", "00C": "CYS", "0EA": "TYR", "F7W": "TRP", "LEI": "VAL", "UMA": "ALA", "OLT": "THR", "4KY": "PRO",
        "MCS": "CYS", "TNQ": "TRP", "HIX": "ALA", "C1X": "LYS", "PAT": "TRP", "T8L": "THR", "DM0": "LYS", "CG6": "CYS",
        "KPF": "LYS", "DYS": "CYS", "BB6": "CYS", "LAL": "ALA", "DLY": "LYS", "DJD": "PHE", "LTU": "TRP", "TYT": "TYR",
        "VPV": "LYS", "D11": "THR", "LEF": "LEU", "1X6": "SER", "ML3": "LYS", "MAA": "ALA", "7ID": "ASP", "AAR": "ARG",
        "NZC": "THR", "R1A": "CYS", "CGV": "CYS", "D3P": "GLY", "TIS": "SER", "LYR": "LYS", "4IN": "TRP", "CY4": "CYS",
        "0AF": "TRP", "TLY": "LYS", "SVA": "SER", "4HH": "SER", "HQA": "ALA", "PHD": "ASP", "KYN": "TRP", "4FW": "TRP",
        "VHF": "GLU", "CTH": "THR", "B3X": "ASN", "MTY": "TYR", "MLY": "LYS", "SMC": "CYS", "TS9": "ILE", "PXU": "PRO",
        "DSE": "SER", "P3Q": "TYR", "BCX": "CYS", "FAK": "LYS", "SVY": "SER", "CSS": "CYS", "FDL": "LYS", "2LT": "TYR",
        "N80": "PRO", "B3A": "ALA", "LYO": "LYS", "VR0": "ARG", "YTH": "THR",
    }

    # Mappings to aa metadata
    _ID_MAP: Dict[int, tuple] = {aa[0]: aa for aa in _AA_LIST}
    _ONE_MAP: Dict[str, tuple] = {aa[1]: aa for aa in _AA_LIST}
    _THREE_MAP: Dict[str, tuple] = {aa[2]: aa for aa in _AA_LIST}
    #_AA_MAP = _ID_MAP | _ONE_MAP | _THREE_MAP

    # Translation tables
    THREE_2_ONE: Dict[str, str] = {aa[2]: aa[1] for aa in _AA_LIST}
    ONE_2_THREE: Dict[str, str] = {aa[1]: aa[2] for aa in _AA_LIST}
    ONE_2_ID: Dict[str, int] = {aa[1]: aa[0] for aa in _AA_LIST}
    ID_2_ONE: Dict[int, str] = {aa[0]: aa[1] for aa in _AA_LIST}

    # Construcors --------------------------------------------------------------
    def __init__(self, aa_one: str) -> "AminoAcid":
        """Only accepts standard Amino Acid one-letter-codes."""
        assert aa_one in self._ONE_MAP, f"ERROR in AminoAcid('{aa_one}'): invalid amino acid one-letter-code."
        aa_metadata = self._ONE_MAP[aa_one]
        self.id: int = aa_metadata[0]
        self.one: str = aa_metadata[1]
        self.three: str = aa_metadata[2]
        self.three_standard: str = aa_metadata[2]
        self.name: str = aa_metadata[3]

    @classmethod
    def parse_three(cls, aa_three: int) -> "AminoAcid":
        """Parse an AminoAcid from its three-letter-code (can handle non-standard AAs and mapping to corresponding standard AA)."""

        # Standard case
        aa_one = cls.THREE_2_ONE.get(aa_three, None)
        if aa_one is not None:
            return AminoAcid(aa_one)
        
        # Non-standard case
        aa_three_standard = cls._NON_STANDARD_AAS.get(aa_three, None)
        if aa_three_standard is not None:
            aa_one = cls.THREE_2_ONE[aa_three_standard]
            aa = AminoAcid(aa_one)
            aa.three = aa_three
            return aa
        
        # Unknown case
        return cls.get_unknown()

    @classmethod
    def get_all(cls) -> List["AminoAcid"]:
        """Get the list of all 20 standard AminoAcids."""
        return [AminoAcid(aa_metadata[1]) for aa_metadata in cls._AA_LIST]
    
    @classmethod
    def get_unknown(cls) -> "AminoAcid":
        """Return an unknown AminoAcid."""
        aa = AminoAcid("A")
        aa.id = AminoAcid.UNK_ID
        aa.one = AminoAcid.UNK_ONE
        aa.three = AminoAcid.UNK_THREE
        aa.three_standard = None
        aa.name = AminoAcid.UNK_NAME
        return aa
    
    @classmethod
    def get_gap(cls) -> "AminoAcid":
        """Return a gap 'AminoAcid'."""
        aa = AminoAcid("A")
        aa.id = AminoAcid.GAP_ID
        aa.one = AminoAcid.GAP_ONE
        aa.three = AminoAcid.GAP_THREE
        aa.three_standard = None
        aa.name = AminoAcid.GAP_NAME
        return aa

    # Base properties ----------------------------------------------------------
    def __str__(self) -> str:
        if self.is_standard:
            return f"AminoAcid('{self.one}', '{self.three}', id={self.id})"
        else:
            return f"AminoAcid('{self.one}', '{self.three}' (std='{self.three_standard}'), id={self.id})"
        
    def is_gap(self) -> bool:
        self.id == AminoAcid.GAP_ID

    def is_unknown(self) -> bool:
        return self.id == AminoAcid.UNK_ID
    
    def is_aminoacid(self) -> bool:
        return not self.is_gap()
    
    def is_standard(self) -> bool:
        return self.three == self.three_standard

    # Class methods ------------------------------------------------------------
    @classmethod
    def id_exists(cls, id: int) -> bool:
        """Return if 'id' corresponds to the id of a standard Amino Acid."""
        return id in cls._ID_MAP

    @classmethod
    def one_exists(cls, aa_one: str) -> bool:
        """Return if 'aa_one' corresponds to the one-letter-code of a standard Amino Acid."""
        return aa_one in cls._ONE_MAP
    
    @classmethod
    def three_exists(cls, aa_three: str) -> bool:
        """Return if 'aa_three' corresponds to the three-letter-code of a standard Amino Acid."""
        return aa_three in cls._THREE_MAP




# Imports ----------------------------------------------------------------------
import os.path
from os import cpu_count
from typing import Union, List, Dict, Literal, Callable
import tempfile
import numpy as np
from rsalor.utils import time_str
from rsalor.sequence import AminoAcid
from rsalor.sequence import Mutation
from rsalor.sequence import Sequence
from rsalor.sequence import FastaReader, FastaStream
from rsalor.sequence import PairwiseAlignment
from rsalor.structure import Structure
from rsalor.weights import compute_weights, read_weights, write_weights
from rsalor.utils import CSV
from rsalor.utils import Logger


# Main -------------------------------------------------------------------------
class MSA:
    """Class MSA: Combines structural data (RSA) and evolutionary data (Log Odd Ratio, LOR from MSA) to evaluate missense mutations in proteins.
    Main class of the RSALOR package.
    """


    # Constants ----------------------------------------------------------------
    ACCEPTED_EXTENTIONS = FastaStream.ACCEPTED_EXTENTIONS
    N_STATES = len(AminoAcid.ONE_2_ID) + 1
    GAP_ID = N_STATES - 1
    GAP_CHAR = AminoAcid.GAP_ONE
    ONE_2_ID = {aa_one: aa_id for aa_one, aa_id in AminoAcid.ONE_2_ID.items()}
    ONE_2_ID_GAP = {aa_one: aa_id for aa_one, aa_id in AminoAcid.ONE_2_ID.items()}
    ONE_2_ID_GAP[GAP_CHAR] = GAP_ID


    # Constructor --------------------------------------------------------------
    def __init__(
            self,
            msa_path: str,
            pdb_path: Union[None, str]=None,
            chain: Union[None, str]=None,
            theta_regularization: float=0.01,
            n_regularization: float=0.0,
            count_target_sequence: bool=True,
            remove_redundant_sequences: bool=True,
            seqid_weights: Union[None, float]=0.80,
            min_seqid: Union[None, float]=0.35,
            num_threads: int=1,
            rsa_solver: Literal["biopython", "DSSP", "MuSiC"]="biopython",
            rsa_solver_path: Union[None, str]=None,
            trimmed_msa_path: Union[None, str]=None,
            allow_msa_overwrite: bool=False,
            weights_cache_path: Union[None, str]=None,
            rsa_cache_path: Union[None, str]=None,
            verbose: bool=False,
            disable_warnings: bool=False,
            name: Union[None, str]=None,
        ):
        """\nRSA*LOR: Combines structural data (RSA) and evolutionary data (Log Odd Ratio, LOR from MSA) to evaluate missense mutations in proteins.
        
    ----------------------------------------------------------------------------
    usage (Python):
      from rsalor import MSA                        # Import pip package
      msa = MSA('./msa1.fasta', './pdb1.pdb', 'A')  # Initialize MSA object with an MSA file, a PDB file and corresponding chain in the PDB
      scores = msa.get_scores()                     # Compute RSA*LOR scores of all single-site missense mutations
      msa.save_scores('./msa1_rsalor-scores.csv')   # Save scores to a '.csv' file

    ----------------------------------------------------------------------------
    Main arguments:
      msa_path (str)                            path to MSA '.fasta' or '.a2m' file (file can be zipped with '.gz')

    Structure arguments:
      pdb_path (None | str, None)               path to PDB '.pdb' file (leave empty to ignore structure)
      chain (None | str, None)                  chain in the PDB to consider
    
    LOR/LR arguments:
      theta_regularization (float, 0.01)        regularization term for LOR/LR at amino acid frequencies level
      n_regularization (float, 0.0)             regularization term for LOR/LR at amino acid counts level
      count_target_sequence (bool, True)        count target (first) sequence of the MSA in frequencies
      remove_redundant_sequences (bool, True)   pre-process MSA to remove redundent sequences
      seqid_weights (None | float, 0.80)        seqid threshold to consider two sequences in the same cluster for weighting (set None to ignore)
      min_seqid (None | float, 0.35)            discard sequences which seqid with target sequence is below (set None to ignore)
      num_threads (int, 1)                      number of threads (CPUs) for weights evaluation (in the C++ backend)

    RSA arguments:
      rsa_solver ('biopython'/'DSSP'/'MuSiC')   used solver to compute RSA (DSSP and MuSiC require the software to be installed)
      rsa_solver_path (None | str, None)        path to DSSP/MuSiC executable to compute RSA (leave empty if software is in system PATH)
      
    Files management arguments:
      trimmed_msa_path (None | str, None)       set to save the trimmed + non-redundent MSA file (leave empty to ignore)
      allow_msa_overwrite (bool, False)         allow to overwrite initial MSA file with the trimmed + non-redundent MSA file
      
    Cache arguments:
      weights_cache_path (None | str, None)     set to read (if file exists) or write (if files does not exists) weights (leave empty to ignore)
      rsa_cache_path (None | str, None)         set to read (if file exists) or write (if files does not exists) rsa values (leave empty to ignore)

    Logging arguments:
      verbose (bool, False)                     log execution steps
      disable_warnings (bool, False)            disable logging of Warnings
      name (None | str, None)                   name of the MSA object (for logging)
        """

        # MSA path Guardians
        self.name = "" # Required for logs, so we set directly.
        self._verify_input_msa_path(msa_path)

        # Fill basic properties
        self.msa_path: str = msa_path
        self.msa_filename: str = os.path.basename(self.msa_path)
        self.name: str = name
        if self.name is None:
            for extention in self.ACCEPTED_EXTENTIONS:
                if self.msa_filename.endswith(f".{extention}"):
                    self.name = self.msa_filename.removesuffix(f".{extention}")
                    break
        self.pdb_path: str = pdb_path
        self.chain: str = chain
        self.rsa_solver: str = rsa_solver
        self.rsa_solver_path: str = rsa_solver_path
        self.rsa_cache_path: str = rsa_cache_path
        self.theta_regularization: float = theta_regularization
        self.n_regularization: float = n_regularization
        self.remove_redundant_sequences: bool = remove_redundant_sequences
        self.count_target_sequence: bool = count_target_sequence
        self.seqid_weights: Union[None, float] = seqid_weights
        self.min_seqid: Union[None, float] = min_seqid
        self.num_threads: int = num_threads
        self.weights_cache_path: str = weights_cache_path
        self.trimmed_msa_path: Union[None, str] = trimmed_msa_path
        self.allow_msa_overwrite: bool = allow_msa_overwrite
        self.verbose: bool = verbose
        self.disable_warnings: bool = disable_warnings
        self.logger = Logger(verbose, disable_warnings, step_prefix="RSALOR", warning_note=f" in {self}", error_note=f" in {self}")

        # Too much CPU warning
        num_cpu_total = cpu_count()
        if num_cpu_total is None or num_cpu_total < self.num_threads:
            self.logger.warning(f"num_threads={num_threads} exeeds total number of CPUs detected on current machine (num_cpu_total={num_cpu_total}).")

        # Init structure (if pdb_path is specified)
        self._init_structure()

        # Read sequences
        self._read_sequences()

        # Filter sequences that are too far from target sequence
        if min_seqid is not None:
            self._remove_far_seqid_sequences()

        # Align Structure and Sequence (if pdb_path is specified)
        self._align_structure_to_sequence()

        # Save trimmed MSA (if trimmed_msa_path is specified)
        if self.trimmed_msa_path is not None:
            self.logger.step("save trimmed MSA (without target sequence gaps and non-std AAs, without redundent sequences) to a file.")
            self.logger.log(f" * trimmed_msa_path: '{trimmed_msa_path}'")
            self._verify_trimmed_seq_path()
            self.write(trimmed_msa_path)

        # Assign weights
        self._init_weights()

        # Counts and Frequencies
        self._init_counts()


    # Constructor dependencies -------------------------------------------------
    def _init_structure(self) -> None:
        """Parse PDB file and compute RSA (Relative Solvent Accessibility)."""

        # Case: pdb_path is None -> just log some warnings and continue
        if self.pdb_path is None:
            if self.chain is not None:
                warning_log = "pdb_path is not set, so structure and RSA are ignored."
                warning_log += f"   However chain is set to '{self.chain}'."
                warning_log += f"   Please specify pdb_path to consider structure and RSA."
                self.logger.warning(warning_log)
            if self.rsa_solver_path is not None:
                warning_log = "pdb_path is not set, so structure and RSA are ignored."
                warning_log += f"   However rsa_solver_path is set to '{self.rsa_solver_path}'."
                warning_log += f"   Please specify pdb_path to consider structure and RSA."
                self.logger.warning(warning_log)
            self.structure = None
            return None
        
        # Set Structure
        self.logger.step(f"parse PDB structure '{os.path.basename(self.pdb_path)}' (chain '{self.chain}') and compute RSA.")
        assert self.chain is not None, f"{self.error_prefix}: pdb_path='{self.pdb_path}' is set, so please set also the PDB chain to consider."
        self.structure = Structure(
            self.pdb_path,
            self.chain,
            rsa_solver=self.rsa_solver,
            rsa_solver_path=self.rsa_solver_path,
            rsa_cache_path=self.rsa_cache_path,
            verbose=self.verbose,
        )

        # Non assigned RSA warnings
        self._verify_rsa_values()

    def _read_sequences(self) -> None:
        """Read sequences from MSA FASTA file."""
        
        # Read MSA
        self.logger.step(f"read sequences from MSA file '{self.msa_filename}'.")

        # Inspect target sequence for gaps and non-standard AAs
        # Also set up alignment between MSA and trimmed MSA positions
        target_sequence = FastaReader.read_first_sequence(self.msa_path)
        self.fasta_to_fasta_trimmed: Dict[str, str] = {}
        self.fasta_trimmed_to_fasta: Dict[str, str] = {}
        tgt_seq_len = len(target_sequence)
        n_gaps = 0
        non_standard = []
        keep_position: List[bool] = []
        i_res_trimmed = 0
        for i_res, res in enumerate(target_sequence):
            if res in self.ONE_2_ID: # Standard AA -> keep
                fasta_res = str(i_res+1)
                fasta_trimmed_res = str(i_res_trimmed+1)
                self.fasta_to_fasta_trimmed[fasta_res] = fasta_trimmed_res
                self.fasta_trimmed_to_fasta[fasta_trimmed_res] = fasta_res
                i_res_trimmed += 1
                keep_position.append(True)
            elif res == self.GAP_CHAR: # Gap -> remove
                n_gaps += 1
                keep_position.append(False)
            else: # Other -> remove
                non_standard.append(res)
                keep_position.append(False)
        n_remove = n_gaps + len(non_standard)
        do_trimming = n_remove > 0
        n_keep = len(target_sequence) - n_remove
        if n_keep < 1:
            raise ValueError(f"{self.error_prefix}: target sequence does not contain any standard amino acid residues.")
        if do_trimming:
            self.logger.warning(f"target sequence contains some gaps or non-standard amino acids: MSA will be trimmed: {len(target_sequence)} -> {n_keep} (num trimmed positions: {n_remove}).")
        if n_gaps > 0:
            self.logger.warning(f"target sequence contains {n_gaps} gaps -> those positions will be trimmed.")
        if len(non_standard) > 0:
            non_std_str = "".join(non_standard)
            if len(non_std_str) > 10:
                non_std_str = non_std_str[0:7] + "..."
            self.logger.warning(f"target sequence contains {len(non_standard)} non-standard amino acids ('{non_std_str}') -> those positions will be trimmed.")

        # Read sequences from file
        self.sequences: List[Sequence] = []
        fasta_stream = FastaStream(self.msa_path) # Caution with this one
        n_tot_sequences = 0
        # Keep redundant sequences
        if not self.remove_redundant_sequences:
            sequence = fasta_stream.get_next()
            while sequence is not None:
                self._verify_sequence_length(sequence, tgt_seq_len, n_tot_sequences)
                if do_trimming:
                    sequence.trim(keep_position)
                if len(sequence) == 0:
                    continue
                self.sequences.append(sequence)
                sequence = fasta_stream.get_next()
                n_tot_sequences += 1
        # Keep only non-redundant sequences
        # the filter is done during execution to optimize time and RAM (could help with huge MSAs)
        else:
            sequences_set = set()
            sequence = fasta_stream.get_next()
            while sequence is not None:
                self._verify_sequence_length(sequence, tgt_seq_len, n_tot_sequences)
                if do_trimming:
                    sequence.trim(keep_position)
                if len(sequence) == 0:
                    continue
                sequence_str = sequence.sequence
                if sequence_str not in sequences_set:
                    self.sequences.append(sequence)
                    sequences_set.add(sequence_str)
                sequence = fasta_stream.get_next()
                n_tot_sequences += 1
            self.logger.log(f" * remove redundant sequences  : {n_tot_sequences} -> {len(self.sequences)}")
        fasta_stream.close()

        # Verify MSA consisency
        assert self.depth > 1, f"{self.error_prefix}: MSA contains no or only 1 sequence."
        assert self.length > 0, f"{self.error_prefix}: MSA target (first) sequence is of length 0."

        # Log
        self.logger.log(f" * MSA length (tgt seq length) : {len(self.target_sequence)}")
        self.logger.log(f" * MSA depth (num sequences)   : {len(self.sequences)}")

        # Set target sequence name
        self.target_sequence.name += " (trimmed MSA)"

    def _remove_far_seqid_sequences(self) -> None:
        """Filter sequences that are too far from target sequence by sequence identity."""

        # Guardian
        assert 0.0 <= self.min_seqid < 1.0, f"{self.error_prefix}: min_seqid={self.min_seqid} should be stricktly between 0 and 1."

        # Log
        self.logger.step(f"filter sequences that are too far from target sequence.")

        # Compute sequences to keep
        keep_sequences: List[Sequence] = []
        target_sequence_str = self.sequences[0].sequence
        for current_sequence in self.sequences:
            current_sequence_str = current_sequence.sequence

            # Compute seqid with target sequence
            current_seqid = self._seqid_to_target(target_sequence_str, current_sequence_str)

            if current_seqid > self.min_seqid:
                keep_sequences.append(current_sequence)

        # Update MSA sequences
        l1, l2 = len(self.sequences), len(keep_sequences)
        self.sequences = keep_sequences

        # Log results
        self.logger.log(f" * filter: {l1} -> {l2} (min_seqid={self.min_seqid:.2f})")

        # Guardians
        if l2 == 0:
            error_log = f"{self.error_prefix}: remove_far_seqid_sequences(): no sequence left."
            error_log += f"\n - No sequences left in the MSA after removing sequences that are too far from target sequence (by sequence indentity)"
            error_log += f"\n - min_seqid={self.min_seqid}: please increase value or set to None."
            raise ValueError(error_log)
        
    def _seqid_to_target(self, seq1: str, seq2: str) -> float:
        """Computes sequence identity between two sequences in the MSA."""
        gap = self.GAP_CHAR
        num_identical_residues = sum([int(aa1 == aa2) for aa1, aa2 in zip(seq1, seq2)])
        num_aligned_residues = sum([int(aa != gap) for aa in seq2])
        return num_identical_residues / num_aligned_residues
        
    def _align_structure_to_sequence(self) -> None:
        """Align residues position between PDB sequence and target sequence of the MSA."""

        # Init
        self.str_seq_align: PairwiseAlignment
        self.pdb_to_fasta_trimmed: Dict[str, str] = {}
        self.fasta_trimmed_to_pdb: Dict[str, str] = {}
        self.rsa_array: List[Union[None, float]] = [None for _ in range(self.length)]
        self.rsa_factor_array: List[Union[None, float]] = [None for _ in range(self.length)]
        if self.structure is None:
            return None
        
        # Log
        self.logger.step("align Structure (from PDB) and Sequence (from MSA).")
        
        # Init alignment
        self.str_seq_align = PairwiseAlignment(self.structure.sequence, self.target_sequence)
        
        # Map positions
        i_pdb, i_fasta_trimmed = 0, 0
        n_no_rsa, n_no_residue = 0, 0
        for aa_pdb, aa_fasta_trimmed in zip(self.str_seq_align.align1, self.str_seq_align.align2):
            if aa_pdb != self.GAP_CHAR and aa_fasta_trimmed != self.GAP_CHAR:
                residue = self.structure.chain_residues[i_pdb]
                fasta_trimmed_id = str(i_fasta_trimmed+1)
                self.pdb_to_fasta_trimmed[residue.resid] = fasta_trimmed_id
                self.fasta_trimmed_to_pdb[fasta_trimmed_id] = residue.resid
                self.rsa_array[i_fasta_trimmed] = residue.rsa
                if residue.rsa is None:
                    n_no_rsa += 1
            if aa_pdb != self.GAP_CHAR:
                i_pdb += 1
            if aa_fasta_trimmed != self.GAP_CHAR:
                if aa_pdb == self.GAP_CHAR: # Position in MSA but not is PDB
                    n_no_residue += 1
                i_fasta_trimmed += 1

        # Log
        n_assigned = len([rsa for rsa in self.rsa_array if rsa is not None])
        self.logger.log(f" * {n_assigned} / {len(self.rsa_array)} assigned RSA values for positions in trimmed MSA")

        # Set RSA factor
        self.set_rsa_factor()
        
        # Alignment Warnings
        if n_no_residue:
            self.logger.warning(f"{n_no_residue} / {len(self.rsa_array)} positions in trimmed MSA with no corresponding residues in PDB structure.")
        if n_no_rsa:
            self.logger.warning(f"{n_no_rsa} / {len(self.rsa_array)} positions in trimmed MSA corresponding to PDB residues without assigned RSA.")
        critical_alignment_warning = False
        if self.str_seq_align.mismatch > 0:
            critical_alignment_warning = True
            self.logger.warning(f"{self.str_seq_align.mismatch} / {len(self.rsa_array)} mismatch between trimmed MSA and PDB.", critical=True)
        if self.str_seq_align.internal_gap2 > 0:
            critical_alignment_warning = True
            self.logger.warning(f"{self.str_seq_align.internal_gap2} internal residues in the PDB do not correspond to a position in trimmed MSA.", critical=True)
        if critical_alignment_warning and not self.disable_warnings:
            self.str_seq_align.show(n_lines=80, only_critical_chunks=True)
            self.logger.warning("Please, make sure the first sequence in your MSA file is the target sequence to mutate.", critical=True)

    def set_rsa_factor(self, rsa_factor_function: Union[Callable[[float], float], None]=None) -> None:

        # Set default function
        if rsa_factor_function is None:
            rsa_factor_function = self.inverse_rsa

        # Log
        self.logger.step(f"set RSA factor (RSA -> w(RSA) with w='{rsa_factor_function.__name__}').")

        # Set RSA factor
        for i, rsa in enumerate(self.rsa_array):
            if rsa is not None:
                self.rsa_factor_array[i] = (1.0 - min(rsa, 100.0) / 100.0)

    def _init_weights(self) -> None:
        """Initialize weights for all sequences of the MSA (using C++ backend or from a cache file)."""

        # Case: keep all weights to 1
        if self.seqid_weights is None:
            # Put weight of first sequence to 0.0 manually to ignore it if required
            if not self.count_target_sequence:
                self.sequences[0].weight = 0.0
            return None

        # Read from cached file case
        if self.weights_cache_path is not None and os.path.isfile(self.weights_cache_path):
            self.logger.step("read weights from cached file.")
            self.logger.log(f" * weights_cache_path: '{self.weights_cache_path}'")
            weights = read_weights(self.weights_cache_path)
            if len(weights) != len(self.sequences):
                error_log = f"{self.error_prefix}: read_weights(weights_cache_path='{self.weights_cache_path}'): "
                error_log += f"\nnumber of parsed weights ({len(weights)}) does not match number of sequences ({len(self.sequences)}) in MSA."
                error_log += f"\n   * Please remove current weights_cache file and re-run weights or set weights_cache_path to None."
                raise ValueError(error_log)
        
        # Re-compute case weights case
        else:
            self.logger.step("compute weights using C++ backend.")
            dt = (0.00000000015 * self.length * self.depth**2) / self.num_threads
            dt_str = time_str(dt)
            self.logger.log(f" * seqid (to compute clusters) : {self.seqid_weights}")
            self.logger.log(f" * expected computation-time   : {dt_str} (with {self.num_threads} CPUs)")

            # Case when processed+trimmed MSA in saved 
            if self.trimmed_msa_path is not None:
                weights = compute_weights(
                    self.trimmed_msa_path,
                    self.length,
                    self.depth,
                    self.seqid_weights,
                    self.count_target_sequence,
                    self.num_threads,
                    self.verbose
                )
            # Case when processed+trimmed MSA is not saved
            else:
                with tempfile.TemporaryDirectory() as tmp_dir:
                    tmp_msa_path = os.path.join(tmp_dir, f"{self.name}.fasta")
                    self.write(tmp_msa_path)
                    weights = compute_weights(
                        tmp_msa_path,
                        self.length,
                        self.depth,
                        self.seqid_weights,
                        self.count_target_sequence,
                        self.num_threads,
                        self.verbose
                    )

            # Verify coherence of computed weights
            if len(weights) != len(self.sequences):
                error_log = f"{self.error_prefix}: compute_weights(): "
                error_log += f"number of computed weights ({len(weights)}) does not match number of sequences ({len(self.sequences)}) in MSA."
                raise ValueError(error_log)

        # Assign weights
        for i, wi in enumerate(weights):
            self.sequences[i].weight = wi

        # Save weights in cache file if required
        if self.weights_cache_path is not None and not os.path.isfile(self.weights_cache_path):
            self.logger.step(f"save computed weights to file '{self.weights_cache_path}'.")
            self.logger.log(f" * weights_cache_path: '{self.weights_cache_path}'")
            write_weights(weights, self.weights_cache_path)
        
    def _init_counts(self) -> None:
        """Initialize residues counts and frequences from the MSA."""

        # Log
        self.logger.step("initialize residues counts and frequencies.")

        # Set Neff
        self.Neff: float = sum([sequence.weight for sequence in self.sequences])
        self.logger.log(f" * Neff (sum of weights): {self.Neff:.2f}")

        # Counts
        self.counts = np.zeros((self.length, self.N_STATES), float)
        for sequence in self.sequences:
            for l, aa in enumerate(sequence):
                aa_id = self.ONE_2_ID.get(aa, self.GAP_ID)
                self.counts[l, aa_id] += sequence.weight
        self.gap_counts = self.counts[:, self.GAP_ID]
        self.nongap_counts = self.Neff - self.gap_counts

        # Frequencies
        self.frequencies = self.counts / self.Neff
        self.gap_frequencies = self.frequencies[:, self.GAP_ID]
        self.nongap_frequencies = 1.0 - self.gap_frequencies

        # CI (Conservation Index)
        self.global_aa_frequencies = np.sum(self.frequencies, axis=0) / self.length
        self.CI = np.sqrt(0.5 * np.sum(((self.frequencies - self.global_aa_frequencies)[:, 0:20])**2, axis=1))

        # Manage regularization and LOR/LR scores
        self.update_regularization(self.theta_regularization, self.n_regularization)

    def update_regularization(self, theta_regularization: float, n_regularization: float) -> "MSA":
        """Update regularization parameters and recompute regularized frequencies.

        Arguments:
        theta_regularization (float):  Regularization at the level of frequencies (add theta to all positional frequencies and normalize)
        n_regularization     (float):  Regularization at the level of counts (add n to all positional counts and normalize)
        """

        # Log
        self.logger.step("compute regularized frequencies.")
        self.logger.log(f" * theta_regularization : {theta_regularization}")
        self.logger.log(f" * n_regularization     : {n_regularization}")

        # Regularization Guardians
        assert theta_regularization >= 0.0, f"{self.error_prefix}: theta_regularization={theta_regularization} should be positive."
        assert n_regularization >= 0.0, f"{self.error_prefix}: n_regularization={n_regularization} sould be positive."
        assert theta_regularization > 0.0 or n_regularization > 0.0, f"{self.error_prefix}: both theta_regularization and n_regularization can not be zero to avoid divering values."

        # Set regularization properties
        self.theta_regularization = theta_regularization
        self.n_regularization = n_regularization

        # Apply n_regularization
        self.frequencies_reg = (self.counts + self.n_regularization) / (self.Neff + (float(self.N_STATES) * self.n_regularization))

        # Apply theta_regularization
        reg_term: float = self.theta_regularization / float(self.N_STATES)
        reg_factor: float = 1.0 - self.theta_regularization
        self.frequencies_reg = reg_factor * self.frequencies_reg + reg_term

        # Compute dependent values
        self.gap_frequencies_reg = self.frequencies_reg[:, self.GAP_ID]
        self.nongap_frequencies_reg = 1.0 - self.gap_frequencies_reg

        # Set LOR and LR
        self.LR = np.log(self.frequencies_reg)
        self.LOR = np.log(self.frequencies_reg / (1.0 - self.frequencies_reg))

        return self


    # Base Properties ----------------------------------------------------------
    @classmethod
    def help(cls) -> None:
        """Log main usage (help) of MSA class in the 'rsalor' package."""
        print(cls.__init__.__doc__)
    
    def __str__(self) -> str:
        return f"MSA('{self.name}')"
    
    def __iter__(self):
        return iter(self.sequences)
    
    def __getitem__(self, id: int) -> str:
        return self.sequences[id]
    
    @property
    def target_sequence(self) -> Sequence:
        return self.sequences[0]

    @property
    def length(self) -> int:
        """Length of each sequence from the MSA."""
        return len(self.target_sequence)

    @property
    def depth(self) -> int:
        """Number of sequences in the MSA."""
        return len(self.sequences)
    
    @property
    def error_prefix(self) -> str:
        """Return error in MSA prefix."""
        return f"\033[91mERROR\033[0m in {self}"
    
    @staticmethod
    def inverse_rsa(rsa_value: float) -> float:
        return 1.0 - min(rsa_value, 100.0) / 100.0

    # Scores (such as LOR) Properties ------------------------------------------
    def get_frequency(self, residue_id: int, amino_acid_one_char: str, regularized: bool=True):
        """Get a given amino acid (regularized) frequency at a given position:

        NOTE: residue_id in FASTA convention (first position is 1) on the trimmed MSA

        Arguments:
        residue_id (int):              position index in fasta convention (first residues is 1)
        amino_acid_one_char (str):     amino acid one-letter-code or gap code '-'
        regularized (bool):            set True for regularized frequencies        
        """
        if regularized:
            return self.frequencies_reg[residue_id - 1, self.ONE_2_ID_GAP[amino_acid_one_char]]
        else:
            return self.frequencies[residue_id - 1, self.ONE_2_ID_GAP[amino_acid_one_char]]
        
    def eval_mutations(
            self,
            mutations_list: List[str],
            mutations_reference: Literal["fasta_trimmed", "fasta", "pdb"]="fasta_trimmed",
            metric: Literal["LOR", "LR"]="LOR",
            use_rsa_factor: bool=False,
            disable_wt_warning: bool=False,
        ) -> List[float]:
        """Return list of LOR (log-add ratio) or LR (log ratio) for each mutation in mutations_list
            * for a mutation: LOR('H13K') = log(freq(H, 13) / 1 - freq(H, 13)) - log(freq(K, 13) / 1 - freq(K, 13))
            * by default, position of the mutation is given in the fasta convention (first residue position is 1) on the trimmed MSA

        NOTE: mutation can be indicated in 3 different references:
            - 'fasta': residues are numbered using the FASTA convention (first residue is 1) using the input MSA target sequence as reference
            - 'fasta_trimmed': residues are numbered using the FASTA convention from the trimmed MSA (without target sequence gaps and non-std AAs)
            - 'pdb': residues are numbered as in the PDB file

        Arguments:
        mutations_list (List[str]):             list of mutations as strings
        mutations_reference (str):              "fasta_trimmed", "fasta", "pdb" to specify which mutation convention to use
        metric (str):                           "LOR" or "LR" to specify which metric to compute
        use_rsa_factor (bool):                  set True to multiply the score by the RSA factor at this position
        disable_wt_warning (bool):              set True to not throw WARNING is mutation wt-aa does not match aa in the target sequence
        """

        # Set metric
        ALLOWED_METRICS = ["LOR", "LR"]
        assert metric in ALLOWED_METRICS, f"{self.error_prefix}.eval_mutations(): metric='{metric}' should be in {ALLOWED_METRICS}."
        if metric == "LOR":
            E_matrix = self.LOR
        else:
            E_matrix = self.LR

        # Uniformize mutations to 'fasta_trimmed' reference
        ALLOWED_MUTATIONS_TYPES = ["fasta_trimmed", "fasta", "pdb"]
        assert mutations_reference in ALLOWED_MUTATIONS_TYPES, f"{self.error_prefix}: mutations_reference='{mutations_reference}' sould be in {ALLOWED_MUTATIONS_TYPES}."
        if mutations_reference == "fasta" or mutations_reference == "pdb":
            residues_map = self.fasta_to_fasta_trimmed if mutations_reference == "fasta" else self.pdb_to_fasta_trimmed
            mutations_list_converted = []
            for mutation in mutations_list:
                wt, resid, mt = mutation[0], mutation[1:-1], mutation[-1]
                if resid not in residues_map:
                    error_log = f"{self.error_prefix}.eval_mutations():"
                    error_log += f"\nMutation '{mutation}' can not be converted from '{mutations_reference}' reference to 'fasta_trimmed' reference."
                    error_log += f"\n - residue '{resid}' may be outside of the range of the MSA"
                    if mutations_reference == "pdb":
                        error_log += f"\n - residue '{resid}' may be missing in the PDB structure"
                    elif mutations_reference == "fasta":
                        error_log += f"\n - residue '{resid}' may be a gap or a non-standard amino acid in the target sequence of initial MSA"
                    raise ValueError(error_log)
                mutation_converted = wt + residues_map[resid] + mt
                mutations_list_converted.append(mutation_converted)
            mutations_list_reference = [Mutation(mut) for mut in mutations_list_converted]
        else:
            mutations_list_reference = [Mutation(mut) for mut in mutations_list]

        # Compute mutations
        dE_arr = []
        for i, mutation in enumerate(mutations_list_reference):
            assert 1 <= mutation.position <= self.length, f"{self.error_prefix}.eval_mutations(): position of mutation='{mutation}' is out of range of target sequence of the MSA."
            if not disable_wt_warning:
                aa_target = self.target_sequence[mutation.position-1]
                aa_mutation = mutation.wt_aa.one
                # Trigger incorrect wt aa warning
                if aa_mutation != aa_target:
                    mutation_description = f"'{mutation}'"
                    if mutations_reference != "fasta_trimmed":
                        mutation_description = f"{mutation_description} ('{mutations_list[i]}' in '{mutations_reference}' reference)"
                    self.logger.warning(f"eval_mutations(): mutation {mutation_description}: wt-aa does not match target sequence aa '{aa_target}'.")
            dE = E_matrix[mutation.position-1, mutation.wt_aa.id] - E_matrix[mutation.position-1, mutation.mt_aa.id]
            dE_arr.append(dE)

        # Modulate by RSA factor
        if use_rsa_factor:
            for i, (mutation, dE) in enumerate(zip(mutations_list_reference, dE_arr)):
                rsa_factor = self.rsa_factor_array[mutation.position-1]
                if rsa_factor is None:
                    dE_arr[i] = None
                else:
                    dE_arr[i] = rsa_factor * dE

        return dE_arr
    
    def get_scores(self, round_digit: Union[None, int]=6, log_results: bool=False,) -> List[dict]:
        """Compute scores (gap_freq, wt_freq, mt_freq, RSA, LOR, RSA*LOR, ...) for each single-site mutation.

        NOTE: mutation are indicated in 3 different references:
            - 'fasta': residues are numbered using the FASTA convention (first residue is 1) using the input MSA target sequence as reference
            - 'fasta_trimmed': residues are numbered using the FASTA convention from the trimmed MSA (without target sequence gaps and non-std AAs)
            - 'pdb': residues are numbered as in the PDB file

        output: List of dictionary with the scores:
            [{mutation_fasta: 'A13G', LOR: 0.4578, ...}, ...]
        """

        # Log
        self.logger.step("compute scores for all single-site mutations.")

        # Compute scores for each single site mutation
        all_aas = AminoAcid.get_all()
        scores = []
        for i, wt in enumerate(self.target_sequence.sequence):
            wt_i = AminoAcid.ONE_2_ID[wt]
            resid_fasta_trimmed = str(i+1)
            resid_fasta = self.fasta_trimmed_to_fasta[resid_fasta_trimmed]
            resid_pdb = self.fasta_trimmed_to_pdb.get(resid_fasta_trimmed, None)
            RSA = self.rsa_array[i]
            RSA_factor = self.rsa_factor_array[i]
            CI = self.CI[i]
            gap_freq = self.gap_frequencies[i]
            wt_freq = self.frequencies[i, wt_i]
            for mt_aa in all_aas:
                mt = mt_aa.one
                mt_i = mt_aa.id
                mutation_fasta_trimmed = wt + resid_fasta_trimmed + mt
                mutation_fasta = wt + resid_fasta + mt
                mutation_pdb = None
                if resid_pdb is not None:
                    mutation_pdb = wt + resid_pdb + mt
                mt_freq = self.frequencies[i, mt_i]
                LOR = self.LOR[i, wt_i] - self.LOR[i, mt_i]
                LR = self.LR[i, wt_i] - self.LR[i, mt_i]
                RSALOR, RSALR = None, None
                if RSA_factor is not None:
                    RSALOR = RSA_factor * LOR
                    RSALR = RSA_factor * LR
                score = {
                    "mutation_fasta": mutation_fasta,
                    "mutation_fasta_trimmed": mutation_fasta_trimmed,
                    "mutation_pdb": mutation_pdb,
                    "gap_freq": gap_freq,
                    "wt_freq": wt_freq,
                    "mt_freq": mt_freq,
                    "CI": CI,
                    "RSA": RSA,
                    "LOR": LOR,
                    "LR": LR,
                    "RSA*LOR": RSALOR,
                    "RSA*LR": RSALR,
                }
                scores.append(score)

        # Round float values if required
        if round_digit is not None:
            for score in scores:
                for prop in ["gap_freq", "wt_freq", "mt_freq", "CI", "RSA", "LOR", "LR", "RSA*LOR", "RSA*LR"]:
                    val = score[prop]
                    if val is not None:
                        score[prop] = round(val, round_digit)

        # Log
        if log_results:
            show_n_entries = 25
            scores_csv = CSV(list(scores[0].keys()), name=self.name)
            scores_csv.add_entries(scores[0:show_n_entries])
            scores_csv.show(n_entries=show_n_entries, max_colsize=23)

        return scores
    
    def save_scores(
            self,
            scores_path: str,
            round_digit: Union[None, int]=None,
            sep: str=";",
            missing_value: Union[None, str]="XXX",
            log_results: bool=False
        ) -> List[dict]:
        """Compute scores (gap_freq, wt_freq, mt_freq, RSA, LOR, RSA*LOR, ...) for each single-site mutation and save it to scores_path as a '.csv' file.

        NOTE: mutation are indicated in 3 different references:
            - 'fasta': residues are numbered using the FASTA convention (first residue is 1) using the input MSA target sequence as reference
            - 'fasta_trimmed': residues are numbered using the FASTA convention from the trimmed MSA (without target sequence gaps and non-std AAs)
            - 'pdb': residues are numbered as in the PDB file

        output: List of dictionary with the scores:
            [{mutation_fasta: 'A13G', LOR: 0.4578, ...}, ...]
        """
        
        # Compute scores
        scores = self.get_scores(round_digit)

        # Log
        self.logger.step("save scores to a file.")
        self.logger.log(f" * scores_path: '{scores_path}'")

        # Format in CSV
        scores_properties = list(scores[0].keys())
        scores_csv = CSV(scores_properties, name=self.name)
        scores_csv.set_sep(sep)
        scores_csv.add_entries(scores)

        # Change None to missing_value
        if missing_value is not None:
            for entry in scores_csv:
                for prop in scores_properties:
                    if entry[prop] is None:
                        entry[prop] = missing_value

        # Log
        if log_results:
            scores_csv.show(n_entries=40, max_colsize=23)

        # Save and return
        if scores_path is not None:
            scores_csv.write(scores_path)
        return scores


    # IO Methods ---------------------------------------------------------------
    def write(self, msa_path: str) -> "MSA":
        """Save MSA to a FASTA MSA file."""

        # Guardians
        msa_path = os.path.abspath(msa_path)
        assert os.path.isdir(os.path.dirname(msa_path)), f"{self.error_prefix}.write(): directory of msa_path='{msa_path}' does not exists."
        assert msa_path.endswith(".fasta"), f"{self.error_prefix}.write(): msa_path='{msa_path}' should end with '.fasta'."

        # Write
        with open(msa_path, "w") as fs:
            fs.write("".join([seq.to_fasta_string() for seq in self.sequences]))
        return self
    

    # Guardians Dependencies ---------------------------------------------------
    # Helpers to verify coherence of inputs and current state

    def _verify_input_msa_path(self, msa_path: str) -> None:
        """For correct format and existance of input msa_path."""

        # Existance
        assert os.path.exists(msa_path), f"{self.error_prefix}: msa_path='{msa_path}' files does not exist."

        # Hint for '.ali' format
        if msa_path.endswith(".ali"):
            error_log = f"{self.error_prefix}: msa_path='{msa_path}' should be in FASTA format."
            error_log += f"\n * msa_path: '{msa_path}'"
            error_log += f"\n * input msa_path is expected to be a MSA file in FASTA ('.fasta') format."
            error_log += f"\n * Please convert the MSA to '.fasta' with python script: "
            error_log += "\nfrom rsalor.utils import ali_to_fasta"
            error_log += "\nali_to_fasta('./my_msa.ali', './my_msa.fasta')\n"
            raise ValueError(error_log)
        
        # ERROR for bad MSA extention
        if not any([msa_path.endswith(f".{ext}") for ext in self.ACCEPTED_EXTENTIONS]):
            error_log = f"{self.error_prefix}: msa_path='{msa_path}' should be in FASTA format (with file extention in {self.ACCEPTED_EXTENTIONS})."
            raise ValueError(error_log)

    def _verify_sequence_length(self, sequence: Sequence, target_length: int, i: int) -> None:
        """For coherence of all sequences in the MSA."""
        if len(sequence) != target_length:
            seq_str = sequence.sequence
            if len(seq_str) > 40:
                seq_str = seq_str[0:37] + "..."
            error_log = f"{self.error_prefix}._read_sequences(): msa_path='{self.msa_path}':"
            error_log += f"\n -> length of sequence [{i+1}] l={len(sequence)} ('{seq_str}') does not match length of target sequence l={target_length}."
            raise ValueError(error_log)

    def _verify_trimmed_seq_path(self) -> None:
        """For coherence of trimmed_msa_path and for safety to not overwrite initial input MSA."""
        trimmed_msa_path = str(os.path.abspath(self.trimmed_msa_path))
        assert os.path.isdir(os.path.dirname(trimmed_msa_path)), f"{self.error_prefix}: directory of trimmed_msa_path='{trimmed_msa_path}' does not exists."
        assert trimmed_msa_path.endswith(".fasta"), f"{self.error_prefix}: trimmed_msa_path='{trimmed_msa_path}' should end with '.fasta'."
        if os.path.normpath(self.msa_path) == os.path.normpath(trimmed_msa_path) and not self.allow_msa_overwrite:
            error_log = f"{self.error_prefix}: trimmed_msa_path='{trimmed_msa_path}' is same as input MSA path."
            error_log == "\nIf trimmed_msa_path is set, the trimmed MSA (without target sequence gaps and non-std AAs) will be saved to this path."
            error_log += "\nWARNING: This operation will overwrite initial input MSA file."
            error_log += "\nTo continue, set argument 'allow_msa_overwrite' to True."
            raise ValueError(error_log)
        
    def _verify_rsa_values(self) -> None:
        """Warnings for non-assigned RSA residues."""
        norsa_std, norsa_non_std = 0, 0
        for residue in self.structure.chain_residues:
            if residue.rsa is None:
                if residue.amino_acid.is_standard():
                    norsa_std += 1
                else:
                    norsa_non_std += 1
        norsa = norsa_std + norsa_non_std
        if norsa > 0:
            warning_log = f"{norsa} / {len(self.structure.chain_residues)} residues with no assigned RSA values ({norsa_std} std and {norsa_non_std} non-std) in PDB target chain '{self.chain}'."
            warning_log += "\n   -> This can be caused by non-standard AAs or missing atoms."
            warning_log += "\n   -> For optimal RSA estimations, we highly recommend to 'repair' the PDB and standardize AAs."
            self.logger.warning(warning_log)




# Logger -----------------------------------------------------------------------
class Logger:

    # Constants ----------------------------------------------------------------
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    # Constructor --------------------------------------------------------------
    def __init__(
            self,
            verbose: bool,
            disable_warnings: bool=True,
            step_prefix: str="STEP",
            warning_prefix: str="WARNING",
            error_prefix: str="ERROR",
            step_note: str="",
            warning_note: str="",
            error_note: str="",
        ):
        """Minimalistic logger:
            * manage verbose and disable_warnings
            * add colored prefixes to logs
        """
        self.verbose = verbose
        self.disable_warnings = disable_warnings
        self._step_prefix = step_prefix
        self._warning_prefix = warning_prefix
        self._error_prefix = error_prefix
        self._step_note = step_note
        self._warning_note = warning_note
        self._error_note = error_note

    # Methods ------------------------------------------------------------------
    @property
    def STEP_PREFIX(self) -> str:
        return f"{self.OKGREEN}{self._step_prefix}{self.ENDC}{self._step_note}"
    
    @property
    def WARNING_PREFIX(self) -> str:
        return f"{self.WARNING}{self._warning_prefix}{self.ENDC}{self._warning_note}"
    
    @property
    def CRITICAL_WARNING_PREFIX(self) -> str:
        return f"{self.FAIL}{self._warning_prefix}{self.ENDC}{self._warning_note}"
    
    @property
    def ERROR_PREFIX(self) -> str:
        return f"{self.FAIL}{self._error_prefix}{self.ENDC}{self._error_note}"

    def log(self, log_str: str) -> None:
        if self.verbose:
            print(log_str)

    def step(self, log_str: str) -> None:
        if self.verbose:
            print(f"{self.STEP_PREFIX}: {log_str}")

    def warning(self, log_str: str, critical: bool=False) -> None:
        prefix = self.WARNING_PREFIX
        if critical:
            prefix = self.CRITICAL_WARNING_PREFIX
        if not self.disable_warnings:
            print(f"{prefix}: {log_str}")

    def error(self, log_str: str) -> None:
        print(f"{self.ERROR_PREFIX}: {log_str}")

    def error_str(self, log_str: str) -> str:
        return f"{self.ERROR_PREFIX}: {log_str}"



# Imports ----------------------------------------------------------------------
import os.path
import sys
from typing import Dict, Union
import tempfile
from contextlib import contextmanager
from Bio.PDB import PDBParser
from Bio.PDB.DSSP import DSSP
from rsalor.utils import find_file
from rsalor.rsa.rsa_solver import RSASolver

# RSADSSP ----------------------------------------------------------------------
class RSADSSP(RSASolver):
    """
    RSADSSP(): Solver for RSA (Relative Solvent Accessibility) from a '.pdb' file using DSSP software.
    
    usage:
        rsa_map = RSADSSP('./soft/DSSP/dssp').run('./my_pdb.pdb')
    """

    # Constants ----------------------------------------------------------------
    CANDIDATES_PATHS = ["mkdssp", "dssp"]
    HELPER_LOG = """-------------------------------------------------------
RSA Solver: DSSP issue:
In order to solve Relative Solvent Accessiblity (RSA), RSALOR package uses:
Python package biopython -> interface with the DSSP algorithms (https://biopython.org/docs/1.75/api/Bio.PDB.DSSP.html).
The DSSP software (free for academic use) has to be installed on your computer.
Please install DSSP (https://swift.cmbi.umcn.nl/gv/dssp/) and specify the path to its executable or add it to system PATH.
DSSP source code can be found here: https://github.com/cmbi/hssp

NOTE: you can still use the RSALOR package without DSSP if you only want LOR values of the MSA without using RSA (just set pdb_path=None).
-------------------------------------------------------"""

    # Methods ------------------------------------------------------------------
    def __str__(self) -> str:
        return "RSASolver['DSSP']"

    def execute_solver(self, pdb_path: str) -> Dict[str, float]:
        """Compute RSA by running DSSP.

        args:
        pdb_path (str):         path to PDB file   

        output:
        {resid: str => RSA: float}     (such as {'A13': 48.57, ...})
        """

        # Init DSSP path (check DSSP executable existance only if software is executed)
        self._init_dssp_path()

        # Run DSSP
        pdb_with_cryst1_line = self._inject_cryst1_line(pdb_path) # Manage CRYST1 line
        # Case: CRYST1 line is already present or this version of DSSP does not requires it
        if pdb_with_cryst1_line is None:
            rsa_map = self._run_dssp_backend(pdb_path)
        # Case: inject CRYST1 line and run DSSP on modified PDB
        else:
            with tempfile.NamedTemporaryFile(delete=True) as temp_file:
                tmp_pdb_path = temp_file.name
                with open(tmp_pdb_path, "w") as fs:
                    fs.write(pdb_with_cryst1_line)
                rsa_map = self._run_dssp_backend(tmp_pdb_path)

        return rsa_map

    # Dependencies -------------------------------------------------------------
    def _init_dssp_path(self) -> str:
        """Find an existing executable file for DSSP on the computer."""
        if self.executable_path is not None:
            dssp_path_list = [self.executable_path] + self.CANDIDATES_PATHS
        else:
            dssp_path_list = self.CANDIDATES_PATHS
        dssp_path = find_file(dssp_path_list, is_software=True, name="DSSP", description=self.HELPER_LOG, verbose=self.verbose)
        self.dssp_path = dssp_path

    def _inject_cryst1_line(self, pdb_path: str) -> Union[None, str]:
        """Inject CRYST1 line in a PDB file if there is not one.
            -> If CRYST1 line is present, return None
            -> Else return a string of the PDB file with the CRYST1 line
        """

        # No need to inject CRYST1 line with mkdssp
        if self.dssp_path.endswith("mkdssp"):
            return None

        # Constants
        CRYST1_HEADER = "CRYST1"
        ATOM_HEADER = "ATOM"
        DEFAULT_CRYST1_LINE = "CRYST1    1.000    1.000    1.000  90.00  90.00  90.00 P 1           1          \n"

        # Read lines
        new_lines = []
        with open(pdb_path, "r") as fs:
            line = fs.readline()

            # Read lines to detect CRYST1 line
            while line:
                if line.startswith(CRYST1_HEADER): # Return None to specify that CRYST1 line is already here
                    return None
                if line.startswith(ATOM_HEADER):
                    new_lines.append(DEFAULT_CRYST1_LINE)
                    new_lines.append(line)
                    line = fs.readline()
                    break
                new_lines.append(line)
                line = fs.readline()
            
            # After injecting CRYST1 line, continue following lines
            while line:
                new_lines.append(line)
                line = fs.readline()

        # Return pdb string with injected CRYST1 line
        return "".join(new_lines)

    def _run_dssp_backend(self, pdb_path: str) -> Dict[str, float]:
        """Run DSSP software with the BioPython interface."""

        # Parse PDB with BioPython
        pdb_name = os.path.basename(pdb_path).removesuffix(".pdb")
        structure = PDBParser(QUIET=True).get_structure(pdb_name, pdb_path)
        model = structure[0]

        # Run DSSP
        if not self.verbose: # Run DSSP with WARNINGS desabled
            with suppress_stderr():
                dssp = DSSP(model, pdb_path, dssp=self.dssp_path)
        else: # Run DSSP normally
            dssp = DSSP(model, pdb_path, dssp=self.dssp_path)

        # Parse Residues
        resid_set = set()
        residues_keys = list(dssp.keys())
        rsa_map: Dict[str, float] = {}
        for res_key in residues_keys:
            chain, (res_insertion, res_id, res_alternate_location) = res_key
            resid = f"{chain}{res_insertion}{res_id}".replace(" ", "")
            if resid not in resid_set:
                res_data = dssp[res_key]
                resid_set.add(resid)
                rsa = res_data[3]
                if isinstance(rsa, float):
                    rsa = round(rsa * 100.0, 4)
                    rsa_map[resid] = rsa

        # Return
        return rsa_map

# Just to delete WARNINGS from DSSP and BioPython ------------------------------
# Because BioPython and DSSP does not provide a disable WARNINGS option ...
@contextmanager
def suppress_stderr():
    """Redirect standard error to null (with some magic)"""
    original_stderr = sys.stderr
    sys.stderr = open(os.devnull, 'w')
    try:
        yield
    finally:
        sys.stderr.close()
        sys.stderr = original_stderr



# Imports ----------------------------------------------------------------------
import os.path
from os import remove
from typing import Union
from Bio import SeqIO

# MSA files processing functions -----------------------------------------------
def ali_to_fasta(input_path: str, output_path: str, delete_input: bool=False) -> Union[None, str]:
    """Convert '.ali' (Stickholm format) file to a '.fasta' file.
        * Then, deletes input '.ali' file if required.
        * Returns output_path or None if execution failed.
    Source: https://stackoverflow.com/questions/24156578/using-bio-seqio-to-write-single-line-fasta
    """

    # Guardians
    error_log = f"ERROR in ali_to_fasta(): "
    error_log += f"\n * input_path  : '{input_path}'"
    error_log += f"\n * output_path : '{output_path}'\n"
    if not os.path.isfile(input_path):
        print(f"{error_log} -> input file does not exists.")
        return None
    if not is_nonempty_file(input_path):
        print(f"{error_log} -> input file is empty.")
        return None
    if not input_path.endswith(".ali"):
        print(f"{error_log} -> input file should end with '.ali'.")
        return None
    if not output_path.endswith(".fasta"):
        print(f"{error_log} -> output file should end with '.fasta'.")
        return None

    # Run convertion
    try:
        records = SeqIO.parse(input_path, "stockholm")
    except Exception as error:
        print(f"{error_log} -> input file parsing failed.")
        print(error)
        return None
    try:
        SeqIO.FastaIO.FastaWriter(output_path, wrap=None).write_file(records)
    except Exception as error:
        print(f"{error_log} -> file convertion + writing failed.")
        print(error)
        return None
    
    # Detect errors
    if not is_nonempty_file(output_path):
        print(f"{error_log} -> converted output file is empty.")
        return None
    
    # Delete initial '.ali' file if required
    if delete_input:
        if os.path.isfile(input_path):
            remove(input_path)

    # Return
    return output_path

# Dependency -------------------------------------------------------------------
def is_nonempty_file(input_path: str) -> bool:
    """Check if 'input_path' is an existing non-empty file."""
    if not os.path.isfile(input_path):
        return False
    with open(input_path, "r") as fs:
        line = fs.readline()
        return len(line) > 0



# Imports ----------------------------------------------------------------------
from rsalor.utils import is_convertable_to
from rsalor.sequence import AminoAcid

# Mutation ---------------------------------------------------------------------
class Mutation:
    """Container class for a single missence/synonymous mutation on a protein (FASTA) sequence.
    
    NOTE: Use FASTA residue position convention: so resdue position is an integer and starts at 1.
    NOTE: Trivial mutations are accepter (like 'A14A').

    usage:
    mutation: Mutation = Mutation('A14G')
    """

    # Constructor --------------------------------------------------------------
    def __init__(self, mutation_str: str):

        # Unpack and guardians
        assert len(mutation_str) >= 3, f"ERROR in Mutation(): invalid mutation_str='{mutation_str}': should be of length 3 or more."
        wt_aa, position, mt_aa = mutation_str[0], mutation_str[1:-1], mutation_str[-1]
        assert AminoAcid.one_exists(wt_aa), f"ERROR in Mutation(): invalid mutation_str='{mutation_str}': wild-type amino acid is incorrect."
        assert AminoAcid.one_exists(mt_aa), f"ERROR in Mutation(): invalid mutation_str='{mutation_str}': mutant amino acid is incorrect."
        assert is_convertable_to(position, int), f"ERROR in Mutation(): invalid mutation_str='{mutation_str}': position must be a stricktly positive integer."
        position = int(position)
        assert position > 0, f"ERROR in Mutation(): invalid mutation_str='{mutation_str}': position must be a stricktly positive integer."

        # Set
        self.wt_aa: AminoAcid = AminoAcid(wt_aa)
        self.position: int = position
        self.mt_aa: AminoAcid = AminoAcid(mt_aa)

    # Methods ------------------------------------------------------------------
    def __str__(self) -> str:
        return f"{self.wt_aa.one}{self.position}{self.mt_aa.one}"
    
    def __int__(self) -> int:
        """Return unique integer code for each mutation."""
        return self.position*10000 + self.wt_aa.id*100 + self.mt_aa.id
    
    def is_trivial(self) -> bool:
        """Return if mutation is trivial (like 'A14A')."""
        return self.wt_aa == self.mt_aa




# Imports ----------------------------------------------------------------------
import os.path
from typing import List, Union
import gzip
from rsalor.sequence import Sequence


# FastaReader ------------------------------------------------------------------ 
class FastaReader:
    """High level FASTA file reader."""

    @classmethod
    def read_first_sequence(cls, fasta_path: str) -> Sequence:
        """Read first sequence from a FASTA file."""
        fasta_stream = FastaStream(fasta_path)
        sequence = fasta_stream.get_next()
        fasta_stream.close()
        return sequence
    
    @classmethod
    def read_sequences(cls, fasta_path: str) -> List[Sequence]:
        """Read all sequences from a FASTA file."""
        fasta_stream = FastaStream(fasta_path)
        sequences = fasta_stream.get_all()
        fasta_stream.close()
        return sequences
    
    @classmethod
    def count_sequences(cls, fasta_path: str) -> int:
        """Count the number of sequences in a FASTA file (just count the '>')."""

        # Guardians
        assert os.path.isfile(fasta_path), f"ERROR in FastaReader.count_sequences(): fasta_path='{fasta_path}' does not exists."
        if not any([fasta_path.endswith(f".{ext}") for ext in FastaStream.ACCEPTED_EXTENTIONS]):
            raise ValueError(f"ERROR in FastaReader.count_sequences(): fasta_path='{fasta_path}' should end with {FastaStream.ACCEPTED_EXTENTIONS}.")

        # Count
        HEADER_START_CHAR = Sequence.HEADER_START_CHAR
        n = 0
        if fasta_path.endswith(".gz"):
            file = gzip.open(fasta_path, "rt") # gzip.open supports text mode ("rt") to read strings (not bytes)
        else:
            file = open(fasta_path, "r")
        line = file.readline()
        while line:
            if line.startswith(HEADER_START_CHAR):
                n += 1
            line = file.readline()
        return n


# FastaStream ------------------------------------------------------------------
class FastaStream:
    """Low level class to stream sequences from a FASTA file (one by one to avoid loading the whole file in RAM).
    
    WARNING: Please use with caution and do not forget to '.close()'.

    usage:
    
    fasta_stream = FastaStream('./fasta/msa1.fasta') \n
    sequence1 = fasta_stream.get_next() \n
    sequence2 = fasta_stream.get_next() \n
    fasta_stream.close()
    """

    # Constants ----------------------------------------------------------------
    ACCEPTED_EXTENTIONS = ["fasta", "a2m", "fasta.gz", "a2m.gz"]
    HEADER_START_CHAR = Sequence.HEADER_START_CHAR

    # Constructor --------------------------------------------------------------
    def __init__(self, fasta_path: str):

        # Guardians
        assert os.path.isfile(fasta_path), f"ERROR in FastaStream(): fasta_path='{fasta_path}' does not exists."
        if not any([fasta_path.endswith(f".{ext}") for ext in self.ACCEPTED_EXTENTIONS]):
            raise  ValueError(f"ERROR in FastaStream(): fasta_path='{fasta_path}' should end with {self.ACCEPTED_EXTENTIONS}.")

        # Init
        self.fasta_path = fasta_path
        if fasta_path.endswith(".gz"):
            self.file = gzip.open(fasta_path, "rt") # gzip.open supports text mode ("rt") to read strings (not bytes)
        else:
            self.file = open(fasta_path, "r")
        self.current_id = -1
        self.current_line = self._next_line()
        
        # First sequence sanity check
        assert self.current_line is not None, f"ERROR in FastaStream(): no sequences found in file '{fasta_path}'."
        assert self._is_current_line_header(), f"ERROR in FastaStream(): first line of file '{fasta_path}' sould be a fasta header (thus start with '{self.HEADER_START_CHAR}').\nline='{self.current_line}'"

    @property
    def is_open(self) -> bool:
        """Return if current file/stream is still open"""
        return self.current_line is not None

    # Methods ------------------------------------------------------------------
    def close(self) -> None:
        """Close file."""
        self.file.close()
        self.current_line = None

    def get_next(self) -> Union[None, Sequence]:
        """Get next Fasta sequence."""
        if self.current_line is None:
            return None
        self.current_id += 1
        header = self.current_line.removesuffix("\n")
        seq_arr = []
        self.current_line = self._next_line()
        while self.current_line:
            if self._is_current_line_header():
                break
            seq_arr.append(self.current_line.removesuffix("\n"))
            self.current_line = self._next_line()

        seq = "".join(seq_arr)
        return Sequence(header, seq)
    
    def get_all(self) -> List[Sequence]:
        """Get all remaining Fasta sequences."""
        fasta_sequence_list = []
        fasta_sequence = self.get_next()
        while fasta_sequence is not None:
            fasta_sequence_list.append(fasta_sequence)
            fasta_sequence = self.get_next()
        return fasta_sequence_list
    
    # Dependencies -------------------------------------------------------------
    def _next_line(self) -> Union[None, str]:
        line = self.file.readline()
        if line == "":
            self.close()
            return None
        return line
    
    def _is_current_line_header(self) -> bool:
        return self.current_line.startswith(Sequence.HEADER_START_CHAR)
